from __future__ import annotations
from datetime import datetime

import hashlib
from typing import Callable, Optional

from psion.oauth2.exceptions import InvalidGrant, InvalidRequest
from psion.oauth2.mixins import AuthorizationCodeMixin, ClientMixin, UserMixin
from psion.webtools import (
    base64url_encode,
    secret_token,
    to_bytes,
    to_string,
    urlencode,
)

from .base import BaseGrant


def plain_challenge(challenge: str, verifier: str):
    return challenge == verifier


def S256_challenge(challenge: str, verifier: str):
    hashed_verifier = to_string(
        base64url_encode(hashlib.sha256(to_bytes(verifier, "ascii")).digest())
    )
    return challenge == hashed_verifier


class AuthorizationCodeGrant(BaseGrant):
    """
    Implementation of the Authorization Code Grant described in the
    `OAuth 2.1 Spec <https://tools.ietf.org/html/draft-parecki-oauth-v2-1#section-4.1>`_.

    In this grant, the client **MUST** obtain an authorization grant from the RO
    and exchange it for an access token. This implementation uses PKCE by default,
    and enforces its use every time.

    This grant allows all types of clients to get grants and tokens.
    Depending on the client, it **MAY** issue refresh tokens as well.

    The PKCE methods supported are `plain` and `S256` and, even though `plain`
    is supported, its usage is **DISCOURAGED**, since it does not provide
    enough security for the flow.
    """

    __response_type__: str = "code"
    __grant_type__: str = "authorization_code"

    _challenges: dict[str, Callable[[str, str], bool]] = {
        "plain": plain_challenge,
        "S256": S256_challenge,
    }

    async def authorize(self, data: dict, client: ClientMixin, user: UserMixin) -> str:
        """
        Validates the `Authorization Request` and construes the `Authorization Response`
        containing the "code" and the "state", if the latter was provided by the Client.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.1.2>`_::

            "code": REQUIRED. The authorization code generated by the
                authorization server. The authorization code MUST expire shortly
                after it is issued to mitigate the risk of leaks. A maximum
                authorization code lifetime of 10 minutes is RECOMMENDED. The
                client MUST NOT use the authorization code more than once. If an
                authorization code is used more than once, the authorization
                server MUST deny the request and SHOULD revoke (when possible) all
                tokens previously issued based on that authorization code. The
                authorization code is bound to the client identifier and redirect
                URI.

            "state": REQUIRED if the "state" parameter was present in the client
                authorization request. The exact value received from the client.

            For example, the authorization server redirects the user-agent by
            sending the following HTTP response:

            HTTP/1.1 302 Found
            Location: https://client.example.com/cb?code=SplxlOBeZQQYbYS6WxSbIA&state=xyz

        :param data: Data of the Authorization Request.
        :type data: dict

        :param client: Client requesting authorization.
        :type client: ClientMixin

        :param user: Currently authenticated User.
        :type user: UserMixin

        :return: URL Encoded Authorization Response data.
        :rtype: str
        """

        data = await self._validate_authorization_request(data)
        self._validate_requested_scopes(data["scopes"], client, data.get("state"))

        code = secret_token(64)
        await self.adapter.save_authorization_code(code, data, client, user)

        response = {"code": code, "state": data.get("state")}

        async for hook in self._execute_hook(
            "authorization_response", code, data, client, user
        ):
            if hook:
                response.update(hook)

        return self._create_authorization_response(data["redirect_uri"], response)

    async def _validate_authorization_request(self, data: dict) -> dict:
        """
        Validates the incoming data from the `Client` to ensure
        that **ALL** the required parameters were provided.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.1.1.3>`_,
        (fields marked in italic are slightly modified
        to fit the Framework's requirements)::

            "response_type": REQUIRED. Value MUST be set to "code".

            "client_id": REQUIRED. The client identifier as described in Section 2.2.

            "code_challenge": REQUIRED. Code challenge.

            "code_challenge_method": *REQUIRED*. Code verifier transformation
                method is "S256" or "plain".

            "redirect_uri": *REQUIRED*. As described in Section 3.1.2.

            "scope": *REQUIRED*. The scope of the access request
                as described by Section 3.3.

            "state": OPTIONAL. An opaque value used by the client to maintain
                state between the request and callback. The authorization server
                includes this value when redirecting the user-agent back to the
                client.

            GET /authorize?response_type=code&client_id=s6BhdRkqt3&state=xyz
                &redirect_uri=https%3A%2F%2Fclient%2Eexample%2Ecom%2Fcb
                &code_challenge=6fdkQaPm51l13DSukcAH3Mdx7_ntecHYd1vi3n0hMZY
                &code_challenge_method=S256 HTTP/1.1
            Host: server.example.com

        :param data: Data of the Authorization Request.
        :type data: dict

        :return: Validated and reformatted Authorization Request data.
        :rtype: dict
        """

        state: Optional[str] = data.get("state")
        code_challenge: str = data.get("code_challenge")
        code_challenge_method: str = data.get("code_challenge_method")
        scope: str = data.get("scope")

        if state:
            if not state or not isinstance(state, str):
                raise InvalidRequest(description='Invalid parameter "state".')

        if not code_challenge or not isinstance(code_challenge, str):
            raise InvalidRequest(description="Invalid code challenge.", state=state)

        if len(code_challenge) < 43 or len(code_challenge) > 128:
            raise InvalidRequest(
                description="The length of the code challenge MUST be between [43, 128] bytes long.",
                state=state,
            )

        if code_challenge_method not in self._challenges.keys():
            raise InvalidRequest(
                description=f'Unknown code challenge "{code_challenge_method}".',
                state=state,
            )

        if not scope or not isinstance(scope, str):
            raise InvalidRequest(description='Invalid parameter "scope".', state=state)

        response = {
            "code_challenge": code_challenge,
            "code_challenge_method": code_challenge_method,
            "redirect_uri": data["redirect_uri"],
            "scopes": scope.split(),
            "state": state,
        }

        async for hook in self._execute_hook("authorization_request", data):
            if hook:
                response.update(hook)

        return response

    async def token(self, data: dict, client: ClientMixin) -> dict:
        """
        Validates the provided `Authorization Code` to check if its data matches
        both the `Client` and the `User`, then issues a new `Access Token` and,
        if allowed, a new `Refresh Token`, with both bound to the `Client` and `User`.

        Whether it succeeds or fails to issue an `Access Token`, the current
        `Authorization Code` **WILL** be deleted at the end of the flow.
        This prevents both `Replay Attacks` and the generation of multiple
        `Access Tokens` against the `Authorization Code`.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-5.1>`_::

            The authorization server issues an access token and optional refresh
            token, and constructs the response by adding the following parameters
            to the entity-body of the HTTP response with a 200 (OK) status code:

            "access_token": REQUIRED. The access token issued by the
                authorization server.

            "token_type": REQUIRED. The type of the token issued as described
                in Section 7.1. Value is case insensitive.

            "expires_in": RECOMMENDED. The lifetime in seconds of the access
                token. For example, the value "3600" denotes that the access
                token will expire in one hour from the time the response was
                generated. If omitted, the authorization server SHOULD provide
                the expiration time via other means or document the default value.

            "refresh_token": OPTIONAL. The refresh token, which can be used to
                obtain new access tokens using the same authorization grant as
                described in Section 6.

            "scope": OPTIONAL, if identical to the scope requested by the
                client; otherwise, REQUIRED. The scope of the access token as
                described by Section 3.3.

            The parameters are included in the entity-body of the HTTP response
            using the "application/json" media type as defined by [RFC7159]. The
            parameters are serialized into a JavaScript Object Notation (JSON)
            structure by adding each parameter at the highest structure level.
            Parameter names and string values are included as JSON strings.
            Numerical values are included as JSON numbers. The order of
            parameters does not matter and can vary.

            The authorization server MUST include the HTTP "Cache-Control"
            response header field [RFC7234] with a value of "no-store" in any
            response containing tokens, credentials, or other sensitive
            information, as well as the "Pragma" response header field [RFC7234]
            with a value of "no-cache".

            For example:

            HTTP/1.1 200 OK
            Content-Type: application/json
            Cache-Control: no-store
            Pragma: no-cache

            {
                "access_token":"2YotnFZFEjr1zCsicMWpAA",
                "token_type":"Bearer",
                "expires_in":3600,
                "refresh_token":"tGzv3JOkF0XG5Qx2TlKWIA",
                "example_parameter":"example_value"
            }

            The client MUST ignore unrecognized value names in the response.  The
            sizes of tokens and other values received from the authorization
            server are left undefined. The client should avoid making
            assumptions about value sizes. The authorization server SHOULD
            document the size of any value it issues.

        :param data: Data of the Token Request.
        :type data: dict

        :param client: Client requesting a new Token.
        :type client: ClientMixin

        :return: Access Token and its metadata, optionally with a Refresh Token.
        :rtype: dict
        """

        try:
            data = await self._validate_token_request(data)
            code = await self.adapter.get_authorization_code(data["code"])

            if not code:
                raise InvalidRequest(description="Invalid Authorization Grant.")

            self._validate_authorization_code(data, code, client)

            user = await self.adapter.find_user(code.get_user_id())

            if not user:
                raise InvalidRequest(description="No user found for this code.")

            access_token = await self.adapter.create_access_token(
                client, user, code.get_scopes()
            )

            refresh_token = (
                await self.adapter.create_refresh_token(client, user, code.get_scopes())
                if "refresh_token" in client.get_grant_types()
                else None
            )

            token = self._create_token(
                access_token,
                self.config.token_lifespan,
                refresh_token,
                code.get_scopes(),
            )

            async for hook in self._execute_hook(
                "token_response", token, data, client, user
            ):
                if hook:
                    token.update(hook)

            return token
        finally:
            # Prevents replay attacks by deleting the code once it is used,
            # whether it succeeds or fails.
            await self.adapter.delete_authorization_code(data["code"])

    async def _validate_token_request(self, data: dict) -> dict:
        """
        Validates the incoming data from the `Client` to ensure
        that **ALL** the required parameters were provided.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.1.3>`_::

            The client makes a request to the token endpoint by sending the
            following parameters using the "application/x-www-form-urlencoded"
            format per Appendix B with a character encoding of UTF-8 in the HTTP
            request entity-body:

            "grant_type": REQUIRED. Value MUST be set to "authorization_code".

            "code": REQUIRED. The authorization code received from the
                authorization server.

            "redirect_uri": REQUIRED, if the "redirect_uri" parameter was
                included in the authorization request as described in
                Section 4.1.1, and their values MUST be identical.

            "client_id": REQUIRED, if the client is not authenticating with the
                authorization server as described in Section 3.2.1.

            "code_verifier": REQUIRED, if the "code_challenge" parameter was
                included in the authorization request. MUST NOT be used
                otherwise. The original code verifier string.

            Confidential or credentialed clients MUST authenticate with the
            authorization server as described in Section 3.2.1.

            For example, the client makes the following HTTP request using TLS
            (with extra line breaks for display purposes only):

            POST /token HTTP/1.1
            Host: server.example.com
            Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW
            Content-Type: application/x-www-form-urlencoded

            grant_type=authorization_code&code=SplxlOBeZQQYbYS6WxSbIA
            &redirect_uri=https%3A%2F%2Fclient%2Eexample%2Ecom%2Fcb
            &code_verifier=3641a2d12d66101249cdf7a79c000c1f8c05d2aafcf14bf146497bed

            The authorization server MUST:

            * require client authentication for confidential and credentialed
                clients (or clients with other authentication requirements),

            * authenticate the client if client authentication is included,

            * ensure that the authorization code was issued to the authenticated
                confidential or credentialed client, or if the client is public,
                ensure that the code was issued to "client_id" in the request,

            * verify that the authorization code is valid,

            * verify that the "code_verifier" parameter is present if and only
                if a "code_challenge" parameter was present in the authorization
                request,

            * if a "code_verifier" is present, verify the "code_verifier" by
                calculating the code challenge from the received "code_verifier"
                and comparing it with the previously associated "code_challenge",
                after first transforming it according to the
                "code_challenge_method" method specified by the client, and

            * ensure that the "redirect_uri" parameter is present if the
                "redirect_uri" parameter was included in the initial authorization
                request as described in Section 4.1.1.3, and if included ensure
                that their values are identical.

        .. note:: The `client_id` is ignored in this flow, since it is already
            used to authenticate the `Client` if needed.

        :param data: Data of the Token Request.
        :type data: dict

        :return: Validated and reformatted Token Request data.
        :rtype: dict
        """

        if not data.get("code") or not isinstance(data.get("code"), str):
            raise InvalidRequest(description='Invalid parameter "code".')

        if not data.get("redirect_uri") or not isinstance(
            data.get("redirect_uri"), str
        ):
            raise InvalidRequest(description='Invalid parameter "redirect_uri".')

        if not data.get("code_verifier") or not isinstance(
            data.get("code_verifier"), str
        ):
            raise InvalidRequest(description='Invalid parameter "code_verifier".')

        response = {
            "code": data["code"],
            "redirect_uri": data["redirect_uri"],
            "code_verifier": data["code_verifier"],
        }

        async for hook in self._execute_hook("token_request", data):
            if hook:
                response.update(hook)

        return response

    def _validate_authorization_code(
        self, data: dict, code: AuthorizationCodeMixin, client: ClientMixin
    ) -> None:
        """
        Validates the data of the `Provided Authorization Code` against both
        the current `Client` and the `Stored Authorization Code`.

        :param data: Provided Authorization Code.
        :type data: AuthorizationCodeModel

        :param code: Stored Authorization Code.
        :type code: AuthorizationCodeMixin

        :param client: Current Client requesting an Access Token.
        :type client: ClientMixin

        :raises InvalidGrant: The data of the Client, the Request
            and the stored Authorization Code do not match.
        :raises InvalidRequest: The transformation method is not supported.
        """

        if data["redirect_uri"] not in client.get_redirect_uris():
            raise InvalidGrant(description="Invalid Redirect URI.")

        if client.get_client_id() != code.get_client_id():
            raise InvalidGrant(description="Mismatching Client ID.")

        if data["redirect_uri"] != code.get_redirect_uri():
            raise InvalidGrant(description="Mismatching Redirect URI.")

        method = self._challenges.get(code.get_code_challenge_method())

        if not method:
            raise InvalidRequest(description=f'Unknown transform "{method}".')

        if not method(code.get_code_challenge(), data["code_verifier"]):
            raise InvalidGrant(description="PKCE challenge failure.")

        if datetime.utcnow() >= code.get_expiration():
            raise InvalidGrant(description="Expired Authorization Code.")

    def _create_authorization_response(self, url: str, data: dict) -> str:
        """
        Creates the Redirect URI with the parameters of the Authorization Response
        to be returned back to the Client.

        :param url: Callback URL of the Client.
        :type url: str

        :param data: Data of the Authorization Response.
        :type data: dict

        :return: URL encoded data of the authorization response.
        :rtype: str
        """

        return urlencode(url, **data)



class Scope(str):
    """
    Defines the model of an OAuth 2.1 Scope.

    The scope is a string that defines a name for Authorization Control.

    This class provides an implementation that exposes both the name of the scope
    as well as a description for documentation and consent purposes.

    Directly accessing the instance returns the name of the scope,
    since the description has no use to the OAuth 2.1 Framework.

    :param name: Name of the scope.
    :type name: str

    :param description: Description of the scope.
    :type description: str
    """

    name: str
    description: str

    def __new__(cls, name: str, description: str = None) -> str:
        klass = super().__new__(cls, name)
        klass.name = name
        klass.description = description
        return klass

    def __init__(self, name: str, description: str = None) -> None:
        pass



from typing import Any, Optional

from psion.webtools import FullDict


class OAuth2Error(Exception):
    """
    Representation of the errors that can occur during the authorization process.

    This is a base class that provides the main attributes defined by
    `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.1.2.1>`_ and
    `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-5.2>`_,
    which are::

        * "error": Denotes the code of the error.
        * "error_description": Human readable description with the details of the error.
        * "error_uri": URI containing more information about the error.
        * "state": String representing the state of the Client before the request.

    :param description: Contains the description of the error.
    :type description: str

    :param uri: Contains the URI that describes the error.
    :type uri: str

    :param state: State of the Client provided in the Request.
    :type state: str, optional

    :param headers: Optional headers attribute mainly used for Basic Authentication.
    :type headers: dict, optional
    """

    error: str = None

    def __init__(
        self,
        description: str = None,
        uri: str = None,
        state: Optional[str] = None,
        headers: Optional[dict] = None,
    ):
        self.description = description
        self.uri = uri
        self.state = state
        self.headers = headers or {}

    def dump(self) -> dict:
        """
        Returns a dictionary representation of the error.

        :return: Error as a dictionary.
        :rtype: dict
        """

        return FullDict(
            error=self.error,
            error_description=self.description,
            error_uri=self.uri,
            state=self.state,
        )


class FatalError(OAuth2Error):
    """
    Returned when the Authorization Endpoint **MUST NOT** redirect the User-Agent
    back to the provided Redirect URI, if any.

    This error **MUST** be re-raised to any of the other errors defined by Guarani,
    since it is not supported by the OAuth 2.1 Spec.
    """

    def __init__(self, error: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.error = error


class InvalidRequest(OAuth2Error):
    error: str = "invalid_request"


class InvalidClient(OAuth2Error):
    error: str = "invalid_client"


class InvalidGrant(OAuth2Error):
    error: str = "invalid_grant"


class UnauthorizedClient(OAuth2Error):
    error: str = "unauthorized_client"


class AccessDenied(OAuth2Error):
    error: str = "access_denied"


class UnsupportedResponseType(OAuth2Error):
    error: str = "unsupported_response_type"


class UnsupportedGrantType(OAuth2Error):
    error: str = "unsupported_grant_type"


class UnsupportedTokenType(OAuth2Error):
    error: str = "unsupported_token_type"


class InvalidScope(OAuth2Error):
    error: str = "invalid_scope"


class ServerError(OAuth2Error):
    error: str = "server_error"


class TemporarilyUnavailable(OAuth2Error):
    error: str = "temporarily_unavailable"



from __future__ import annotations

from datetime import datetime

from psion.jose.exceptions import ExpiredToken, InvalidJWTClaim, NotYetValidToken
from psion.jose.jwk import JsonWebKey
from psion.jose.jws import JsonWebSignature, JsonWebSignatureHeader
from psion.webtools import FullDict, json_dumps, json_loads, to_bytes, to_string


class JsonWebTokenClaims(dict):
    """
    Implementation of RFC 7519.

    It provides validation for the default parameters of the JWT claims.

    The JWT Claims is a JSON object that contains information about
    an application, system or user.

    :param claims: Dictionary containing the parameters
        of the payload of the JWT.
    :type claims: dict

    :param options: Options used to validate the claims.
    :type options: dict[str, dict]
    """

    def __init__(self, claims: dict, options: dict[str, dict] = None) -> None:
        if not claims or not isinstance(claims, dict):
            raise InvalidJWTClaim("The claims object MUST be a valid dict.")

        self.options = options

        self._validate_default_claims(claims)

        if isinstance(options, dict):
            self._validate_claims()

        super().__init__(FullDict(claims))

    def _validate_default_claims(self, claims: dict) -> None:
        # Used to validate the date attributes.
        now = int(datetime.utcnow().timestamp())

        # Validate aud
        if aud := claims.get("aud"):
            if aud is not None and not isinstance(aud, (str, list)):
                raise InvalidJWTClaim('Invalid claim "aud".')

            if isinstance(aud, list):
                if any(not isinstance(item, str) for item in aud):
                    raise InvalidJWTClaim('Invalid claim "aud".')

        # Validate exp
        if exp := claims.get("exp"):
            if exp is not None and type(exp) is not int:
                raise InvalidJWTClaim('Invalid claim "exp".')

            if now >= exp:
                raise ExpiredToken

        # Validate iat
        if iat := claims.get("iat"):
            if iat is not None and type(iat) is not int:
                raise InvalidJWTClaim('Invalid claim "iat".')

            if now < iat:
                raise InvalidJWTClaim('Invalid claim "iat".')

        # Validate iss
        if claims.get("iss") is not None and not isinstance(claims.get("iss"), str):
            raise InvalidJWTClaim('Invalid claim "iss".')

        # Validate jti
        if claims.get("jti") is not None and not isinstance(claims.get("jti"), str):
            raise InvalidJWTClaim('Invalid claim "jti".')

        # Validate nbf
        if nbf := claims.get("nbf"):
            if nbf is not None and type(nbf) is not int:
                raise InvalidJWTClaim('Invalid claim "nbf".')

            if now < nbf:
                raise NotYetValidToken

        # Validate sub
        if claims.get("sub") is not None and not isinstance(claims.get("sub"), str):
            raise InvalidJWTClaim('Invalid claim "sub".')

    def _validate_claims(self) -> None:
        """
        Validates the provided claims using the declared options.

        :raises InvalidJWTClaim: The requested claim does not meet the requirements.
        """

        for claim, option in self.options.items():
            if not isinstance(option, dict):
                continue

            value = self.get(claim)

            if option.get("essential") and value is None:
                raise InvalidJWTClaim(f'Missing required claim "{claim}".')

            if option.get("value") and value != option.get("value"):
                raise InvalidJWTClaim(
                    f'Mismatching expected value "{option.get("value")}". Got "{value}".'
                )

            if option.get("values"):
                if not isinstance(option["values"], list):
                    raise InvalidJWTClaim('Expected a list for the attribute "values".')

                if isinstance(value, list) and any(
                    item in option["values"] for item in value
                ):
                    break

                for expected_value in option["values"]:
                    if value == expected_value:
                        break
                else:
                    raise InvalidJWTClaim(
                        f'Mismatching any of expected values {option["values"]}. Got "{value}".'
                    )


class JsonWebToken:
    """
    Implementation of RFC 7519.

    The JWT is used for transporting claims over the network,
    providing a signature that guarantees the integrity of the information received.

    This implementation provides a set of attributes (described below) to represent
    the state of the information, as well as segregating the header from the payload,
    which in turn facilitates the use of any of them.

    It can be used with either a JWS or a JWE. The most common way of representing
    a JWT is through the JWS Compact Serialization, which gives a small token
    that is digitally signed.

    The claims are represented via a JSON object that contains information about
    an application, system or user. Since this information is digitally signed,
    the receiver can then use the respective key to validate the token and can
    trust that the information is legit.

    TODO: Add support for JWE headers.

    :param claims: Claims about the entity represented by the token.
    :type claims: dict

    :param header: Dictionary that comprise the header of the token.
    :type header: dict

    :param options: Optional validation options for the claims of the token.
    :type options: dict[str, dict]
    """

    def __init__(self, claims: JsonWebTokenClaims, header: dict):
        self.header = JsonWebSignatureHeader(header)
        self.claims = JsonWebTokenClaims(claims)
        self._jws = JsonWebSignature(to_bytes(json_dumps(self.claims)), self.header)

    def __repr__(self):
        return f"<Header: {self.header}, Claims: {self.claims}>"

    def encode(self, key: JsonWebKey) -> str:
        """
        Encodes the internal representation of the current JWT object,
        signs it with the provided key and returns the respective token.

        :param key: Key used to sign and encode the token.
        :type key: JsonWebKey

        :return: Encoded Json Web Token header, payload and signature.
        :rtype: bytes
        """

        return to_string(self._jws.serialize(key))

    @classmethod
    def decode(
        cls,
        token: bytes | str,
        key: JsonWebKey,
        algorithm: str = None,
        validate: bool = True,
        options: dict[str, dict] = None,
    ) -> JsonWebToken:
        """
        Decodes a token checking if its signature matches its content.

        Despite being optional, it is recommended to provide an algorithm
        to prevent the "none attack" and the misuse of a public key
        as secret key.

        The algorithm specified at the header of the token **MUST** match
        the provided algorithm, if any.

        If the token has an Issued At (`iat`) parameter, it will verify the
        validity of the token against the provided `expiration` argument.

        :param token: Token to be verified.
        :type token: Union[bytes, str]

        :param key: Key used to validate the token's signature.
        :type key: JsonWebKey

        :param algorithm: Expected algorithm of the token, defaults to None.
        :type algorithm: str, optional

        :param validate: Defines if the decoding should validate the signature.
            Defaults to True.
        :type validate: bool, optional

        :param options: Optional validation options for the claims of the token.
        :type options: dict[str, dict]

        :return: Instance of a JsonWebToken.
        :rtype: JsonWebToken
        """

        jws = JsonWebSignature.deserialize(to_bytes(token), key, algorithm, validate)
        claims = JsonWebTokenClaims(
            json_loads(jws.payload), options if validate else None
        )

        return cls(claims, jws.header)



import abc
from typing import Awaitable, Callable

from psion.oauth2.mixins import ClientMixin
from psion.oauth2.models import Request


class BaseAuthentication(abc.ABC):
    """
    Base class for handling client authentication on an arbitrary endpoint.

    This package implements the default authentication methods described
    by the `token_endpoint_auth_method` metadata specified in
    `RFC 7591 <https://tools.ietf.org/html/rfc7591#section-2/>`_.

    To add custom client authentication flows, the application MUST inherit
    this base class and implement its abstract method :meth:`authenticate`.

    :cvar `__method__`: The name of the authentication method.

    :param find_client: Callable used to retrieve a Client
        from the application's storage.
    :type find_client: Callable[[str], Awaitable[ClientMixin]]
    """

    __method__: str = None

    def __init__(self, find_client: Callable[[str], Awaitable[ClientMixin]]) -> None:
        self.find_client = find_client

    @abc.abstractmethod
    async def authenticate(self, request: Request) -> ClientMixin:
        """
        Gets the client from the application's storage and validates its data.

        :param request: Current request being processed.
        :type request: Request

        :raises InvalidClient: The requested client is invalid.

        :return: Authenticated Client.
        :rtype: ClientMixin
        """



class JoseError(Exception):
    """
    Base error class for the exceptions of the JOSE implementation.

    :cvar error: Default error message of the exception.
    """

    error: str = None

    def __init__(self, error: str = None) -> None:
        super().__init__(error or self.error)


class ExpiredToken(JoseError):
    error = "The provided Json Web Token is expired."


class InvalidJWSHeader(JoseError):
    error: str = "The provided Json Web Signature Header is invalid."


class InvalidJWSSerialization(JoseError):
    error = "The provided JWS Serialization is invalid."


class InvalidJWTClaim(JoseError):
    error = "The provided Json Web Key Token contains an invalid claim."


class InvalidKey(JoseError):
    error: str = "The provided key is invalid or contains invalid parameters."


class InvalidKeySet(JoseError):
    error: str = "The provided key set is invalid or contain invalid keys."


class InvalidSignature(JoseError):
    error: str = "The provided signature does not match the provided data."


class NotYetValidToken(JoseError):
    error = "The provided Json Web Token is not yet valid."


class UnsupportedAlgorithm(JoseError):
    error = "The provided algorithm is not supported."


class UnsupportedParsingMethod(JoseError):
    error: str = "The provided parsing method is not supported."



from __future__ import annotations

import abc
from typing import Any

from psion.jose import JsonWebKeySet
from psion.oauth2.adapter import BaseAdapter
from psion.oauth2.authentication import ClientAuthentication
from psion.oauth2.config import Config
from psion.oauth2.endpoints import (
    AuthorizationEndpoint,
    RevocationEndpoint,
    TokenEndpoint,
)
from psion.oauth2.grants import BaseGrant
from psion.oauth2.models import Request, Response, Scope


class BaseProvider(abc.ABC):
    def __init__(
        self,
        issuer: str,
        *,
        adapter: type[BaseAdapter],
        grants: list[type[BaseGrant]],
        scopes: list[Scope],
        keyset: JsonWebKeySet,
        error_url: str,
        token_lifespan: int = 3600,
        id_token_lifespan: int = 36000,
    ) -> None:
        self.issuer = issuer
        self.scopes = scopes

        self.adapter = adapter()
        self.config = Config(
            issuer=issuer,
            scopes=scopes,
            token_lifespan=token_lifespan,
            id_token_lifespan=id_token_lifespan,
            keyset=keyset,
            error_url=error_url,
        )

        self.authenticate = ClientAuthentication(self.adapter.find_client)
        self.grants = [grant(self.adapter, self.config) for grant in grants]

        self.authorize = AuthorizationEndpoint(
            grants=self.grants,
            adapter=self.adapter,
            config=self.config,
            authenticate=self.authenticate,
        )

        self.revoke = RevocationEndpoint(
            adapter=self.adapter,
            config=self.config,
            authenticate=self.authenticate,
        )

        self.token = TokenEndpoint(
            grants=self.grants,
            adapter=self.adapter,
            config=self.config,
            authenticate=self.authenticate,
        )

    def add_hook(self, grant_cls: type[BaseGrant], hook: Any) -> None:
        """
        Adds a hook to the specified Grant, extending its functionalities.

        :param grant: Grant to be extended.
        :type grant: BaseGrant

        :param hook: Class containing the desired hooks' implementations.
        :type hook: Any
        """

        for grant in self.grants:
            if not isinstance(grant, grant_cls):
                continue

            for checkpoint in grant.__hooks__:
                if hasattr(hook, checkpoint):
                    grant.__hooks__[checkpoint].add(hook(self.adapter, self.config))

    @abc.abstractmethod
    async def create_request(self, request: Any) -> Request:
        """
        Transforms the Web Server's request into an Request object.

        This method **MUST** be implemented in integrations.

        :param request: Web Server's specific request object.
        :type request: Any

        :return: Transformed request object.
        :rtype: Request
        """

    @abc.abstractmethod
    async def create_response(self, response: Response) -> Any:
        """
        Transforms the `Response` object into a Response of the integrated Web Server.

        This method **MUST** be implemented in integrations.

        :param response: Framework's Response.
        :type response: Response

        :return: Integrated Web Server Response.
        :rtype: Any
        """



from __future__ import annotations

import copy
from typing import Any
from urllib.parse import quote_plus, urlparse

from psion.webtools import FullDict, json_dumps, to_bytes, to_string, urldecode

from psion.oauth2.mixins import ClientMixin, UserMixin


class Request:
    """
    Implementation of the OAuth 2.1 Request.

    It has roughly the same attributes and methods of a request of a web framework.

    It is provided as a framework-agnostic version of the request to allow
    for multiple integrations without breaking functionality.

    :param method: HTTP Method of the current request.
    :type method: str

    :param url: URL of the current request.
    :type url: str

    :param headers: Headers of the current request.
    :type headers: dict

    :param body: Body of the current request, defaults to None.
    :type body: bytes, optional

    :param user: Currently authenticated user, defaults to None.
    :type user: Any, optional

    :ivar method: Method of the current request. Only `GET` and `POST` are supported.
    :ivar url: Full URL of the current request.
    :ivar path: Path string of the URL.
    :ivar query: Query parameters of the URL.
    :ivar fragment: Fragment parameters of the URL.
    :ivar headers: Headers of the current request.
    :ivar user: Currently authenticated user.
    :ivar client: Client that made the OAuth 2.1 request, resolved on execution.
    :ivar data: Data of the current request. Parses the query and then the form.
    """

    def __init__(
        self,
        method: str,
        url: str,
        headers: dict[str, Any],
        body: bytes = None,
        user: UserMixin = None,
    ) -> None:
        if method.lower() not in ("get", "post"):
            raise RuntimeError(f'The method "{method}" is not supported.')

        self.method = method
        self.url = url
        self.path = urlparse(url).path
        self.query = urldecode(urlparse(url).query)
        self.fragment = urldecode(urlparse(url).fragment)
        self.headers = {k.lower(): v for k, v in headers.items()}
        self.user: UserMixin = user
        self.client: ClientMixin = None

        self._body = body

        # Data of the current request.
        self.data = copy.deepcopy(self.query)
        self.data.update(self.form())

    def form(self) -> dict:
        """
        Parses the body as **application/x-www-form-encoded**
        and returns a dictionary of the parameters of the form.

        :return: Dictionary of the parsed body.
        :rtype: dict
        """

        return urldecode(to_string(self._body))


class Response:
    """
    Implementation of the OAuth 2.1 Response.

    It has roughly the same attributes and methods of a response of a web framework.

    It is provided as a framework-agnostic version of the response to allow
    for multiple integrations without breaking functionality.

    :cvar media_type: Media type of the response.

    :param status_code: HTTP Status Code of the Response, defaults to 200.
    :type status_code: int

    :param headers: HTTP Headers of the Response, defaults to None.
    :type headers: dict, optional

    :param body: HTTP Body of the Response, defaults to None.
    :type body: Any, optional
    """

    media_type: str = None

    def __init__(
        self, status_code: int = 200, headers: dict = None, body: Any = None
    ) -> None:
        self.status_code = status_code
        self.headers = self.parse_headers(headers)
        self.body = self.parse_body(body)

    def parse_headers(self, headers: dict) -> dict:
        """
        Parses the headers of the response.

        If no header is provided, it returns an empty dictionary.

        :param headers: Headers of the response.
        :type headers: dict

        :return: Parsed headers with the optional media type of the response.
        :rtype: dict
        """

        if headers is None:
            headers = {}

        return FullDict({"Content-Type": self.media_type}, **headers)

    def parse_body(self, body: Any = None) -> bytes:
        """
        Parses the body of the response.

        If no body is provided, it returns an empty byte-string.

        :param body: Object representing the body of the response, defaults to None.
        :type body: Any, optional

        :return: Parsed body as a byte-string.
        :rtype: bytes
        """

        return to_bytes(body) or b""


class JSONResponse(Response):
    """
    Representation of a JSON Response.

    The body of the Response **MUST** be JSON Encodable.

    :cvar media_type: `application/json`
    """

    media_type: str = "application/json"

    def parse_body(self, body: Any = None) -> bytes:
        """
        Returns a bytes representation of the JSON Encodable body of the Response.

        :param body: JSON Encodable object representing the body of the response,
            defaults to None.
        :type body: Any, optional

        :return: Byte-string of the JSON representation of the Response Body.
        :rtype: bytes
        """

        return to_bytes(json_dumps(body))


class RedirectResponse(Response):
    """
    Representation of a Redirect Response.

    It will return a redirect-ready object to the provided URL.
    No body object is to be provided.

    :param url: URL to be redirected.
    :type url: str

    :param status_code: HTTP Status Code of the Redirection, defaults to 303.
    :type status_code: int, optional

    :param headers: HTTP Headers of the Redirection, defaults to None.
    :type headers: dict, optional
    """

    def __init__(self, url: str, status_code: int = 303, headers: dict = None) -> None:
        super().__init__(status_code=status_code, headers=headers, body=b"")
        self.headers["Location"] = quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")



from __future__ import annotations

from dataclasses import dataclass

from psion.jose.jwk import JsonWebKeySet

from psion.oauth2.models import Scope


@dataclass
class Config:
    issuer: str
    scopes: list[Scope]
    token_lifespan: int
    id_token_lifespan: int
    keyset: JsonWebKeySet
    error_url: str



from psion.oauth2.exceptions import InvalidClient
from psion.oauth2.mixins import ClientMixin
from psion.oauth2.models import Request

from .base import BaseAuthentication


class None_(BaseAuthentication):
    """
    Implements the Client Authentication via the Body of the Request.

    If this workflow is enabled, it will look at the body of the request
    for a scheme similar to the following::

        client_id=client1

    The request's body often comes with more information that may pertain to
    a specific endpoint or authorization grant. In this case,
    the body will be similar to the following::

        key1=value1&key2=value2&client_id=client1

    In this workflow, if the client provides a client_secret,
    it will automatically fail, since it is intended to be used by public clients.
    """

    __method__: str = "none"

    async def authenticate(self, request: Request) -> ClientMixin:
        body = request.form()

        client_id, client_secret = body.get("client_id"), body.get("client_secret")

        if not client_id or client_secret is not None:
            return None

        client = await self.find_client(client_id)

        if not client:
            raise InvalidClient("Client not found.")

        if client.get_client_secret() != client_secret:
            raise InvalidClient("Mismatching Client Secret.")

        if client.get_token_endpoint_auth_method() != self.__method__:
            raise InvalidClient(
                f'This Client is not allowed to use the method "{self.__method__}".'
            )

        return client



from __future__ import annotations

import abc
from collections.abc import AsyncGenerator
from typing import Any, Optional

from psion.oauth2.adapter import BaseAdapter
from psion.oauth2.config import Config
from psion.oauth2.exceptions import InvalidScope
from psion.oauth2.mixins import ClientMixin, UserMixin
from psion.webtools import FullDict


class BaseGrant:
    """
    Base class responsible for defining the interface of an OAuth 2.1 Grant.

    The `Grant` is responsible for generating `Access Tokens` and, in some cases,
    `Authorization Grants` for Clients that need access to a protected resource
    of a web application (REST API, Web App, GraphQL, etc).

    Clients that want to access those protected resources **MUST** first obtain
    the consent of the `Resource Owner` for the desired resources. The resources
    can be accessed if the authorization contains the required `Scope` for it.

    Once the `Resource Owner` has granted the authorization to the `Client`,
    this one can then obtain an `Access Token` from the `Authorization Server`.

    This `Access Token` contains the scopes granted by the `Resource Owner` and
    allows the client to access the resources without further interaction
    with the `Resource Owner`.

    It is important to note that the scopes are defined on the application level,
    since they reflect its resources and the `Access Policies` of the application.

    :cvar ``__authentication_methods__``: Allowed Client Authentication methods.
    :cvar ``__grant_type__``: Name of the Token Grant.
    :cvar ``__response_type__``: Name of the Authorization Grant.
    :cvar ``__hooks__``: Defines the hooks to be executed in the grant flow
        to provide custom functionalities to it.

    :param adapter: Instance of the adapter used by the application.
    :type adapter: BaseAdapter

    :param config: Configuration of the provider.
    :type config: Configuration
    """

    __authentication_methods__: list[str] = None
    __grant_type__: str = None
    __response_type__: str = None
    __hooks__: dict[str, set[Any]] = {
        "authorization_request": set(),
        "authorization_response": set(),
        "token_request": set(),
        "token_response": set(),
    }

    def __init__(self, adapter: BaseAdapter, config: Config):
        self.adapter = adapter
        self.config = config

    def _create_token(
        self,
        access_token: str,
        expires_in: int,
        refresh_token: Optional[str] = None,
        scopes: Optional[list[str]] = None,
    ) -> dict:
        """
        Creates a `Token Response` with the following format::

            {
                "access_token": "2YotnFZFEjr1zCsicMWpAA",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token":"tGzv3JOkF0XG5Qx2TlKWIA",
                "scope": "scope1 scope2"
            }

        :param access_token: Access Token issued to the Client.
        :type access_token: str

        :param expires_in: Lifespan of the Access Token in seconds.
        :type expires_in: int

        :param refresh_token: Optional Refresh Token issued to the Client,
            defaults to None.
        :type refresh_token: str, optional

        :param scopes: Optional list of Scopes if the granted scopes are different
            than the scopes requested.
        :type scopes: list[str], optional

        :return: Bearer Token Response with the provided parameters.
        :rtype: dict
        """

        return FullDict(
            {
                "access_token": access_token,
                "token_type": "Bearer",
                "expires_in": expires_in,
                "refresh_token": refresh_token,
                "scope": " ".join(scopes) if scopes else None,
            }
        )

    async def _execute_hook(
        self, name: str, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[dict]:
        """
        Executes all the registered hooks that contain the requested method name.

        :param name: Name of the checkpoint of the Grant flow.
        :type name: str

        :return: Async generator containing the results from the executed hooks.
        :rtype: AsyncGenerator[dict]
        """

        for checkpoint, hooks in self.__hooks__.items():
            if checkpoint == name:
                for hook in hooks:
                    method = getattr(hook, name)
                    yield await method(*args, **kwargs)

    def _validate_requested_scopes(
        self, scopes: list[str], client: ClientMixin, state: Optional[str] = None
    ) -> None:
        """
        Verifies if all of the requested `Scopes`
        are supported by the `Authorization Server`.

        :param scopes: Requested scopes.
        :type scopes: list[str]

        :param client: Client that requested the provided scopes.
        :type client: ClientMixin

        :param state: State of the request provided by the Client, defaults to None.
        :type state: str, optional

        :raises InvalidScope: The Authorization Server does not support
            one or more of the requested scopes.
        """

        if scopes is None:
            return

        for scope in scopes:
            if scope not in self.config.scopes:
                raise InvalidScope(
                    description=f'Unsupported scope "{scope}".', state=state
                )

        if not set(scopes).issubset(set(client.get_allowed_scopes(scopes))):
            raise InvalidScope(
                description="This client is not authorized to request this scope.",
                state=state,
            )

    @abc.abstractmethod
    async def authorize(self, data: dict, client: ClientMixin, user: UserMixin) -> str:
        """
        Validates the data provided by the `Client` in the `Authorization`
        portion of the `Grant` and creates the `Authorization Response`
        that will be returned to the `Client`.

        :param data: Data received from the Client's Authorization Request.
        :type data: dict

        :param client: Client requesting authorization.
        :type client: ClientMixin

        :param user: Currently authenticated user granting authorization to the Client.
        :type user: UserMixin

        :return: Encoded URL containing the response from the Authorization Grant.
        :rtype: str
        """

    @abc.abstractmethod
    async def token(self, data: dict, client: ClientMixin) -> dict:
        """
        Validates the data provided by the `Client` in the `Token`
        portion of the `Grant` and creates the `Token Response`
        that will be returned to the `Client`.

        :param data: Data received from the Client's Authorization Request.
        :type data: dict

        :param client: Client requesting authorization.
        :type client: ClientMixin

        :return: Dictionary containing the response from the Token Grant.
        :rtype: dict
        """



from __future__ import annotations

import inspect
from typing import Callable
from psion.jose.exceptions import (
    InvalidJWSHeader,
    InvalidJWSSerialization,
    InvalidKey,
    UnsupportedAlgorithm,
)

from psion.jose.jwa.jws import (
    JWSAlgorithm,
    none,
    HS256,
    HS384,
    HS512,
    RS256,
    RS384,
    RS512,
    ES256,
    ES384,
    ES512,
    PS256,
    PS384,
    PS512,
)
from psion.jose.jwk import JsonWebKey
from psion.webtools import (
    FullDict,
    base64url_decode,
    base64url_encode,
    json_dumps,
    json_loads,
)


class JsonWebSignatureHeader(dict):
    """
    Implementation of RFC 7515.

    This is the implementation of the Header of the Json Web Signature.
    It provides validation for the default parameters of the JWS header.

    The JWS Header is a JSON object that provides information on how to
    manipulate the payload of the message, such as permitted algorithms
    and the keys to be used in signing and verifying the payload.

    TODO: Add support for RFC 7797.

    :cvar ``__algorithms__``: Algorithms supported by the JWS.

    :param header: Dictionary containing the parameters of the JWS header.
    :type header: dict
    """

    __algorithms__: dict[str, JWSAlgorithm] = {
        none.__algorithm__: none,
        HS256.__algorithm__: HS256,
        HS384.__algorithm__: HS384,
        HS512.__algorithm__: HS512,
        RS256.__algorithm__: RS256,
        RS384.__algorithm__: RS384,
        RS512.__algorithm__: RS512,
        ES256.__algorithm__: ES256,
        ES384.__algorithm__: ES384,
        ES512.__algorithm__: ES512,
        PS256.__algorithm__: PS256,
        PS384.__algorithm__: PS384,
        PS512.__algorithm__: PS512,
    }

    def __init__(self, header: dict):
        if not header or not isinstance(header, dict):
            raise InvalidJWSHeader

        validators: list[Callable[[dict], None]] = [
            method
            for name, method in inspect.getmembers(self, predicate=inspect.ismethod)
            if name.startswith("validate_")
        ]

        for validator in validators:
            validator(header)

        if (alg := header.get("alg")) not in self.__algorithms__.keys():
            raise InvalidJWSHeader(f'Unknown JWS algorithm "{alg}".')

        self.algorithm = self.__algorithms__[alg]
        super().__init__(FullDict(header))

    def validate_alg(self, header: dict) -> None:
        """
        The alg parameter is mandatory, and **MUST** be a registered algorithm.

        :param header: Dictionary with the JWS Header parameters.
        :type header: dict

        :raises InvalidJWSHeader: Did not find "alg".
        :raises UnsupportedAlgorithm: The algorithm is not supported.
        """

        if "alg" not in header.keys():
            raise InvalidJWSHeader('Missing parameter "alg".')

        if (alg := header.get("alg")) not in self.__algorithms__.keys():
            raise UnsupportedAlgorithm(f'Unsupported algorithm "{alg}".')

    def validate_jku(self, header: dict) -> None:
        pass

    def validate_jwk(self, header: dict) -> None:
        pass

    def validate_kid(self, header: dict) -> None:
        """
        ID of the JWK used by this JWS. If present, MUST be a string.

        :param header: Dictionary with the JWS Header parameters.
        :type header: dict

        :raises InvalidJWSHeader: Key ID is invalid.
        """

        if "kid" in header.keys():
            # pylint: disable=E0601
            if not (kid := header.get("kid")) or not isinstance(kid, str):
                raise InvalidJWSHeader('Invalid parameter "kid".')

    def validate_x5u(self, header: dict):
        pass

    def validate_x5c(self, header: dict):
        pass

    def validate_x5t(self, header: dict):
        pass

    def validate_x5tS256(self, header: dict):
        pass

    def validate_typ(self, header: dict):
        """
        Type of the JWS. If present, MUST be a string.

        :param header: Dictionary with the JWS Header parameters.
        :type header: dict

        :raises InvalidJWSHeader: The type is not a string.
        """

        if "typ" in header.keys():
            # pylint: disable=E0601
            if not (typ := header.get("typ")) or not isinstance(typ, str):
                raise InvalidJWSHeader('Invalid parameter "typ".')

    def validate_cty(self, header: dict):
        """
        Type of the payload. If present, MUST be a string.

        :param header: Dictionary with the JWS Header parameters.
        :type header: dict

        :raises InvalidJWSHeader: Content Type is not a string.
        """

        if "cty" in header.keys():
            # pylint: disable=E0601
            if not (cty := header.get("cty")) or not isinstance(cty, str):
                raise InvalidJWSHeader('Invalid parameter "cty".')

    def validate_crit(self, header: dict):
        """
        Critical parameters of the JWS header. If present, MUST be a list of strings.

        :param header: Dictionary with the JWS Header parameters.
        :type header: dict

        :raises InvalidJWSHeader: Not a list of strings,
            or the critical parameter was not found in the JWS header.
        """

        if "crit" in header.keys():
            # Ensures the type safety of the parameter.
            # pylint: disable=E0601
            if not (crit := header.get("crit")) or not isinstance(crit, list):
                raise InvalidJWSHeader('Invalid parameter "crit".')

            # Ensures that each critical parameter is a VALID string.
            if any(not item or not isinstance(item, str) for item in crit):
                raise InvalidJWSHeader('Invalid parameter "crit".')

            for item in crit:
                if not header.get(item):
                    raise InvalidJWSHeader(f'The parameter "{item}" is required.')


class JsonWebSignature:
    """
    Implementation of RFC 7515.

    The JWS is used for transporting data on the network, providing a signature
    that guarantees the integrity of the information received.

    This implementation provides a set of attributes (described below) to represent
    the state of the information, as well as segregating the header from the payload,
    which in turn facilitates the use of any of them.

    It provides an algorithm attribute as well. The algorithm is used to sign
    and verify the data of the JWS.

    TODO: Add JSON Serialization.

    :ivar header: The header of the JWS.
    :ivar payload: The data that is being handled.

    :param payload: Data to be used by the JWS. MUST be a bytes sequence.
    :type payload: bytes

    :param header: Dictionary that comprise the JWS Header.
    :type header: JsonWebSignatureHeader
    """

    def __init__(self, payload: bytes, header: JsonWebSignatureHeader):
        if not isinstance(header, JsonWebSignatureHeader):
            header = JsonWebSignatureHeader(header)

        self.header = header
        self.payload = payload

    def serialize(self, key: JsonWebKey) -> bytes:
        """
        Serializes the content of the current JsonWebSignature.

        It serializes the header into a Base64Url version of its JSON representation,
        and serializes the payload into a Base64Url format, allowing the compatibility
        of the payload in different systems.

        It creates a byte string message of the following format::

            Base64Url(UTF-8(header)).Base64Url(payload)

        It then signs the message using the provided key, and imbues the signature
        into the message, resulting in the following token::

            Base64Url(UTF-8(header)).Base64Url(payload).Base64Url(signature)

        The above token is then returned to the application.

        :param key: Key used to sign the message.
        :type key: JsonWebKey

        :raises InvalidKey: The provided key is invalid.

        :return: Signed token representing the content of the JWS.
        :rtype: bytes
        """

        _validate_key(key, self.header)

        message = b".".join(
            [base64url_encode(json_dumps(self.header)), base64url_encode(self.payload)]
        )

        signature = self.header.algorithm.sign(message, key)
        return b".".join([message, signature])

    @classmethod
    def deserialize(
        cls,
        token: bytes,
        key: JsonWebKey,
        algorithm: str = None,
        validate: bool = True,
    ) -> JsonWebSignature:
        """
        Deserializes a token checking if its signature matches its content.

        Despite being optional, it is recommended to provide an algorithm
        to prevent the "none attack" and the misuse of a public key
        as secret key.

        The algorithm specified at the header of the token
        **MUST** match the provided algorithm, if any.

        :param token: Token to be deserialized.
        :type token: bytes

        :param key: Key used to validate the token's signature.
        :type key: JsonWebKey

        :param algorithm: Expected algorithm of the token.
        :type algorithm: str

        :param validate: Defines if the deserialization should validate the signature.
            Defaults to True.
        :type validate: bool, optional

        :raises InvalidJWSSerialization: The provided JWS token is invalid.
        :raises InvalidKey: The provided key is invalid.
        :raises InvalidSignature: Unmatching token signature and content.

        :return: Instance of a JsonWebSignature.
        :rtype: JsonWebSignature
        """

        try:
            header, payload, signature = token.split(b".")
        except (AttributeError, ValueError):
            raise InvalidJWSSerialization

        jws_header = JsonWebSignatureHeader(json_loads(base64url_decode(header)))

        if algorithm:
            if (alg := jws_header.get("alg")) != algorithm:
                raise InvalidJWSSerialization(
                    f'Expected algorithm "{algorithm}". Got "{alg}".'
                )

        if validate:
            _validate_key(key, jws_header)
            jws_header.algorithm.verify(signature, b".".join([header, payload]), key)

        return cls(base64url_decode(payload), jws_header)


def _validate_key(key: JsonWebKey, header: JsonWebSignatureHeader):
    """
    Validates the provided key against the header
    algorithm's specifications and restrictions.

    :param key: JWK to be validated.
    :type key: JsonWebKey

    :param header: JWS Header used to validate the key against.
    :type header: JsonWebSignatureHeader

    :raises InvalidKey: The provided key is invalid.
    """

    if not isinstance(key, JsonWebKey):
        raise InvalidKey

    if key.data.get("alg"):
        if key.data.get("alg") != header.get("alg"):
            raise InvalidKey(
                f'This key cannot be used by the algorithm "{header.get("alg")}".'
            )

    if header.get("kid"):
        if key.data.get("kid") != header.get("kid"):
            raise InvalidKey("The key ID does not match the specified on the header.")

    if key.data.get("use"):
        if key.data.get("use") != "sig":
            raise InvalidKey("This key cannot be used to sign a JWS.")

    if key.data.get("key_ops"):
        if any(op not in ("sign", "verify") for op in key.data.get("key_ops")):
            raise InvalidKey("This key cannot be used to sign a JWS.")



from starlette.requests import Request as StarletteRequest
from starlette.responses import Response as StarletteResponse

from psion.oauth2.models import Request, Response

from .base import BaseProvider


class StarletteProvider(BaseProvider):
    async def create_request(self, request: StarletteRequest) -> Request:
        return Request(
            method=request.method,
            url=str(request.url),
            headers=dict(request.headers),
            body=await request.body(),
            user=request.user,
        )

    async def create_response(self, response: Response) -> StarletteResponse:
        return StarletteResponse(response.body, response.status_code, response.headers)



from __future__ import annotations

from datetime import datetime
from typing import Optional

from psion.oauth2.exceptions import InvalidClient, InvalidRequest, InvalidScope
from psion.oauth2.mixins import ClientMixin, RefreshTokenMixin

from .base import BaseGrant


class RefreshTokenGrant(BaseGrant):
    """
    Implementation of the `Refresh Token` grant described in the
    `OAuth 2.1 Spec <https://tools.ietf.org/html/draft-parecki-oauth-v2-1#section-6>`_.

    This grant is used by `Clients` with delegated access to get a new `Access Token`
    without the need of a new interaction with the `Resource Owner` during the
    duration of the refresh token.

    Since it issues new access tokens to clients with delegated access, it accepts
    all the client types supported by the OAuth 2.1 Spec.
    """

    __grant_type__: str = "refresh_token"

    def _validate_requested_scopes(
        self,
        token: RefreshTokenMixin,
        scopes: list[str],
        client: ClientMixin,
        state: Optional[str] = None,
    ) -> list[str]:
        """
        Verifies if all of the requested `Scopes` are supported
        by the `Authorization Server` and if the requested scopes
        are not broader than previously granted.

        :param token: Stored Refresh Token.
        :type token: RefreshTokenMixin

        :param scopes: Requested scopes.
        :type scopes: list[str]

        :param client: Client that requested the provided scopes.
        :type client: ClientMixin

        :param state: State of the request provided by the Client, defaults to None.
        :type state: str, optional

        :raises InvalidScope: The Authorization Server does not support
            one or more of the requested scopes, or the requested scopes
            are broader than previously granted.

        :return: Scopes of the new Token.
        :rtype: list[str]
        """

        super()._validate_requested_scopes(scopes, client, state)

        if not scopes:
            return token.get_scopes()

        if any(scope not in token.get_scopes() for scope in scopes):
            raise InvalidScope(description="Invalid broader scope.")

        return list(set(token.get_scopes()).intersection(scopes))

    async def token(self, data: dict, client: ClientMixin) -> dict:
        """
        Validates the provided `Refresh Token` to check if its data matches
        both the `Client` and the `User`, then issues a new `Access Token` and,
        if allowed, a new `Refresh Token`, with both bound to the `Client` and `User`.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-5.1>`_::

            The authorization server issues an access token and optional refresh
            token, and constructs the response by adding the following parameters
            to the entity-body of the HTTP response with a 200 (OK) status code:

            "access_token": REQUIRED. The access token issued by the
                authorization server.

            "token_type": REQUIRED. The type of the token issued as described
                in Section 7.1. Value is case insensitive.

            "expires_in": RECOMMENDED. The lifetime in seconds of the access
                token. For example, the value "3600" denotes that the access
                token will expire in one hour from the time the response was
                generated. If omitted, the authorization server SHOULD provide
                the expiration time via other means or document the default value.

            "refresh_token": OPTIONAL. The refresh token, which can be used to
                obtain new access tokens using the same authorization grant as
                described in Section 6.

            "scope": OPTIONAL, if identical to the scope requested by the
                client; otherwise, REQUIRED. The scope of the access token as
                described by Section 3.3.

            The parameters are included in the entity-body of the HTTP response
            using the "application/json" media type as defined by [RFC7159]. The
            parameters are serialized into a JavaScript Object Notation (JSON)
            structure by adding each parameter at the highest structure level.
            Parameter names and string values are included as JSON strings.
            Numerical values are included as JSON numbers. The order of
            parameters does not matter and can vary.

            The authorization server MUST include the HTTP "Cache-Control"
            response header field [RFC7234] with a value of "no-store" in any
            response containing tokens, credentials, or other sensitive
            information, as well as the "Pragma" response header field [RFC7234]
            with a value of "no-cache".

            For example:

            HTTP/1.1 200 OK
            Content-Type: application/json
            Cache-Control: no-store
            Pragma: no-cache

            {
                "access_token": "2YotnFZFEjr1zCsicMWpAA",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token": "tGzv3JOkF0XG5Qx2TlKWIA",
                "example_parameter": "example_value"
            }

            The client MUST ignore unrecognized value names in the response.  The
            sizes of tokens and other values received from the authorization
            server are left undefined. The client should avoid making
            assumptions about value sizes. The authorization server SHOULD
            document the size of any value it issues.

        :param data: Data of the Token Request.
        :type data: dict

        :param client: Client requesting a new Token.
        :type client: ClientMixin

        :return: Access Token and its metadata, optionally with a Refresh Token.
        :rtype: dict
        """

        data = self._validate_token_request(data)

        token = await self.adapter.get_refresh_token(data["refresh_token"])

        if not token:
            raise InvalidRequest(description="Invalid Refresh Token.")

        self._validate_refresh_token(token, client)

        scopes = self._validate_requested_scopes(token, data["scopes"], client)
        scopes = client.get_allowed_scopes(scopes)

        user = await self.adapter.find_user(token.get_user_id())

        access_token = await self.adapter.create_access_token(client, user, scopes)
        refresh_token = (
            await self.adapter.create_refresh_token(client, user, scopes)
            if "refresh_token" in client.get_grant_types()
            else None
        )

        return self._create_token(
            access_token,
            self.config.token_lifespan,
            refresh_token,
            scopes if scopes != token.get_scopes() else None,
        )

    def _validate_token_request(self, data: dict) -> dict:
        """
        Validates the incoming data from the `Client` to ensure
        that **ALL** the required parameters were provided.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-6>`_::

            Authorization servers SHOULD determine, based on a risk assessment,
            whether to issue refresh tokens to a certain client. If the
            authorization server decides not to issue refresh tokens, the client
            MAY refresh access tokens by utilizing other grant types, such as the
            authorization code grant type. In such a case, the authorization
            server may utilize cookies and persistent grants to optimize the user
            experience.

            If refresh tokens are issued, those refresh tokens MUST be bound to
            the scope and resource servers as consented by the resource owner.
            This is to prevent privilege escalation by the legitimate client and
            reduce the impact of refresh token leakage.

            If the authorization server issued a refresh token to the client, the
            client makes a refresh request to the token endpoint by adding the
            following parameters using the "application/x-www-form-urlencoded"
            format per Appendix B with a character encoding of UTF-8 in the HTTP
            request entity-body:

            "grant_type": REQUIRED. Value MUST be set to "refresh_token".

            "refresh_token": REQUIRED. The refresh token issued to the client.

            "scope": OPTIONAL. The scope of the access request as described by
                Section 3.3. The requested scope MUST NOT include any scope not
                originally granted by the resource owner, and if omitted is
                treated as equal to the scope originally granted by the resource
                owner.

            Because refresh tokens are typically long-lasting credentials used to
            request additional access tokens, the refresh token is bound to the
            client to which it was issued. Confidential or credentialed clients
            MUST authenticate with the authorization server as described in
            Section 3.2.1.

            For example, the client makes the following HTTP request using
            transport-layer security (with extra line breaks for display purposes
            only):

            POST /token HTTP/1.1
            Host: server.example.com
            Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW
            Content-Type: application/x-www-form-urlencoded

            grant_type=refresh_token&refresh_token=tGzv3JOkF0XG5Qx2TlKWIA

            The authorization server MUST:

            * require client authentication for confidential or credentialed
                clients

            * authenticate the client if client authentication is included and
                ensure that the refresh token was issued to the authenticated
                client, and

            * validate the refresh token.

        :param data: Data of the Token Request.
        :type data: dict

        :return: Validated and reformatted Token Request data.
        :rtype: dict
        """

        if not data.get("refresh_token") or not isinstance(
            data.get("refresh_token"), str
        ):
            raise InvalidRequest(description='Invalid parameter "refresh_token".')

        if data.get("scope"):
            if not data.get("scope") or not isinstance(data.get("scope"), str):
                raise InvalidRequest(description='Invalid parameter "scope".')

        return {
            "refresh_token": data["refresh_token"],
            "scopes": data.get("scope").split() if data.get("scope") else None,
        }

    def _validate_refresh_token(self, token: RefreshTokenMixin, client: ClientMixin):
        """
        Validates the data of the `Provided Refresh Token` against both
        the current `Client` and the `Stored Refresh Token`.

        :param code: Stored Refresh Token.
        :type code: RefreshTokenMixin

        :param client: Current Client requesting an Access Token.
        :type client: ClientMixin
        """

        if token.get_client_id() != client.get_client_id():
            raise InvalidClient

        if datetime.utcnow() >= token.get_expiration():
            raise InvalidRequest(description="Refresh Token expired.")



from __future__ import annotations

from typing import Awaitable, Callable

from psion.oauth2.exceptions import InvalidClient
from psion.oauth2.mixins import ClientMixin
from psion.oauth2.models import Request

from .methods import BaseAuthentication


class ClientAuthentication:
    """
    Class that loads all the authentication methods implemented by Guarani
    and the custom methods implemented by the application.

    To authenticate a request, simply invoke the instance as a callable
    passing in the current request and, optionally, the allowed methods
    for the endpoint.

    If the Client uses more than one authentication method on the request,
    or if the Client is not allowed to use the requested method, this class
    will raise an InvalidClient error.

    :param find_client: Function used to search a Client on the application's storage.
    :type find_client: Callable[[str], Awaitable[ClientMixin]]
    """

    def __init__(self, find_client: Callable[[str], Awaitable[ClientMixin]]) -> None:
        self.auth_methods: list[BaseAuthentication] = [
            method(find_client) for method in BaseAuthentication.__subclasses__()
        ]

    async def __call__(
        self, request: Request, methods: list[str] = None
    ) -> ClientMixin:
        """
        Gets the client from the application's storage and validates its data.

        :param request: Current request being handled.
        :type request: Request

        :param methods: Methods allowed by the endpoint, defaults to None.
            If no value is provided, it tests against all registered methods.
        :type methods: list[str], optional

        :raises InvalidClient: The requested client is invalid.

        :return: Authenticated Client.
        :rtype: ClientMixin
        """

        for method in self.auth_methods:
            if methods and method.__method__ not in methods:
                continue

            client = await method.authenticate(request)

            if not client:
                continue

            request.client = client
            return client

        raise InvalidClient



from __future__ import annotations

import base64
import binascii
import copy
import hashlib
import json
import secrets
import string
import struct
from functools import reduce
from operator import concat
from typing import Any, Union
from urllib.parse import parse_qsl, urlencode as urlencode_


class FullDict(dict):
    """
    Dictionary without None values.
    """

    def __init__(self, data: dict = None, /, **kwargs: Any):
        _data = copy.deepcopy(data or {})
        _data.update(kwargs)
        super().__init__({k: v for k, v in _data.items() if v is not None})

    def __setitem__(self, key: Any, value: Any):
        if value is not None:
            super().__setitem__(key, value)


def to_bytes(data: Any, encoding: str = "utf-8") -> bytes:
    """
    Returns the bytes representation of the provided data.

    :param data: Data to be transformed.
    :type data: Any

    :param encoding: Encoding of the bytes encoding, defaults to "utf-8".
    :type encoding: str, optional

    :return: Bytes representation of the provided data.
    :rtype: bytes
    """

    if isinstance(data, bytes) or data is None:
        return data

    if isinstance(data, str):
        return data.encode(encoding)

    if isinstance(data, (int, float)):
        return str(data).encode(encoding)

    return bytes(data, encoding=encoding)


def to_string(data: Any, encoding: str = "utf-8") -> str:
    """
    Returns the string representation of the provided data.

    :param data: Data to be transformed.
    :type data: Any

    :param encoding: Encoding of the string encoding, defaults to "utf-8".
    :type encoding: str, optional

    :return: String representation of the provided data.
    :rtype: str
    """

    if isinstance(data, str) or data is None:
        return data

    if isinstance(data, bytes):
        try:
            return data.decode(encoding)
        except (TypeError, UnicodeDecodeError):
            return base64url_encode(data).decode(encoding)

    return str(data, encoding=encoding)


def base64url_decode(data: bytes) -> bytes:
    """
    Decodes a URL Safe Base64 encoded bytes string into its original contents.

    :param data: Data to be decoded.
    :type data: bytes

    :return: Original contents of the provided encoded data.
    :rtype: bytes
    """

    data += b"=" * (len(data) % 4)
    return base64.urlsafe_b64decode(data)


def base64url_encode(data: Any) -> bytes:
    """
    Returns a URL Safe Base64 encoding representation of the provided data.

    :param data: Data to be encoded.
    :type data: Any

    :return: URL Safe Base64 encoded representation of the provided data.
    :rtype: bytes
    """

    return base64.urlsafe_b64encode(to_bytes(data)).rstrip(b"=")


def b64_to_int(data: Union[bytes, str]) -> int:
    """
    Decodes a URL Safe Base64 representation of an integer.

    :param data: Data to be decoded.
    :type data: Union[bytes, str]

    :raises TypeError: The provided data is not a valid URL Safe Base64 string.

    :return: Decoded Integer.
    :rtype: int
    """

    if data is None:
        return data

    if not isinstance(data, (bytes, str)):
        raise TypeError('The argument "data" MUST be a bytes or str object.')

    x = base64url_decode(to_bytes(data, "ascii"))
    buffer = struct.unpack("%sB" % len(x), x)

    return int("".join(["%02x" % byte for byte in buffer]), 16)


def int_to_b64(data: int) -> bytes:
    """
    Encodes an integer into a URL Safe Base64 bytes string.

    :param data: Integer to be encoded.
    :type data: int

    :raises ValueError: The data is not an integer.
    :raises ValueError: The data is not a natural number.

    :return: URL Safe Base64 encoded version of the integer.
    :rtype: bytes
    """

    if not isinstance(data, int):
        raise ValueError("Must be a natural number.")

    if data < 0:
        raise ValueError("Must be a natural number.")

    res = data.to_bytes((data.bit_length() + 7) // 8, "big", signed=False)
    return base64url_encode(res)


def json_dumps(data: dict) -> str:
    """
    Dumps a dictionary into a formated JSON string.

    :param data: Dictionary to be formatted.
    :type data: dict

    :return: JSON formatted string of the data.
    :rtype: str
    """

    return json.dumps(data, ensure_ascii=False)


def json_loads(data: Union[bytes, str]) -> dict:
    """
    Loads a JSON string into a dictionary.

    :param data: JSON string to be loaded.
    :type data: Union[bytes, str]

    :raises TypeError: The data is not a valid string.

    :return: Dictionary of the loaded JSON string.
    :rtype: dict
    """

    if not isinstance(data, (bytes, str)):
        raise TypeError("The data MUST be either a string or a bytes object.")

    return json.loads(to_string(data))


def secret_token(size: int = 32) -> str:
    """
    Generates a cryptographically secure, urlsafe random token based on the size.

    :param size: Size of the token, defaults to 32.
    :type size: int, optional

    :return: Token generated.
    :rtype: str
    """

    alphabet = f"{string.ascii_letters}{string.digits}-_"
    return reduce(concat, (secrets.choice(alphabet) for _ in range(size)))


def urlencode(url: str, **params) -> str:
    """
    Encodes keyword arguments into a x-www-form-urlencoded
    string and concatenates it into the provided url.

    :param url: Base URL.
    :type url: str

    :return: Encoded URL containing the parameters of the dictionary as a query string.
    :rtype: str
    """

    return f"{url}?{urlencode_(FullDict(params))}"


def urldecode(query_string: str) -> dict:
    """
    Decodes a x-www-form-urlencoded string into a dictionary.

    :param query_string: Query string to be decoded.
    :type query_string: str

    :return: Dictionary representation of the query string.
    :rtype: dict
    """

    return dict(parse_qsl(query_string))


def get_basic_authorization(headers: dict) -> tuple[str, str]:
    """
    Extracts the authentication credentials from a Basic authentication scheme.

    :param headers: Dictionary of the headers.
    :type headers: dict

    :return: Authentication credentials.
    :rtype: tuple[str, str]
    """

    auth: str = headers.get("authorization") or headers.get("Authorization")

    if not auth or not isinstance(auth, str) or " " not in auth:
        return None, None

    method, token = auth.split(None, 1)

    if method.lower() != "basic":
        return None, None

    try:
        credentials = to_string(base64.b64decode(token))
    except (binascii.Error, TypeError):
        return None, None

    if ":" not in credentials:
        return None, None

    try:
        client_id, client_secret = credentials.split(":", 1)
    except ValueError:
        return None, None

    return client_id, client_secret


def get_bearer_authorization(headers: dict) -> str:
    """
    Extracts a Bearer token from a Bearer authorization scheme.

    :param headers: Dictionary of the headers.
    :type headers: dict

    :return: Bearer token.
    :rtype: str
    """

    auth: str = headers.get("authorization") or headers.get("Authorization")

    if not auth or not isinstance(auth, str) or " " not in auth:
        return None

    method, token = auth.split(None, 1)

    if method.lower() != "bearer":
        return None

    return token


def create_half_hash(access_token: str, alg: str) -> str:
    """
    Creates a Base64UrlEncoded strinf of the left-most
    half of the hash of the Access Token.

    :param access_token: Access Token issued to the Client.
    :type access_token: str

    :param alg: Algorithm used to sign the Access Token.
    :type alg: str

    :return: Base64UrlEncoded half hash of the Access Token.
    :rtype: str
    """

    method = getattr(hashlib, f"sha{alg[2:]}")

    hash_bytes = method(to_bytes(access_token)).digest()
    half_length = int(len(hash_bytes) / 2)
    half_hash = base64url_encode(hash_bytes[:half_length])

    return to_string(half_hash)



import abc

from psion.oauth2.adapter import BaseAdapter
from psion.oauth2.authentication import ClientAuthentication
from psion.oauth2.config import Config
from psion.oauth2.models import Request, Response


class BaseEndpoint(abc.ABC):
    """
    Base class for the endpoints of the OAuth 2.1 framework and its extensions.

    Any endpoint being implemented by the application or by extensions **MUST**
    inherit from this class and implement its abstract methods.

    The type, status, headers and body of the response it returns,
    as well as its meaning and formatting have to be documented
    by the respective endpoint.

    The method :meth:`create_response` **MUST NOT** raise exceptions.
    It **MUST** catch the exceptions and return a valid error response instead.

    :cvar `__endpoint__`: Name of the endpoint.

    :param adapter: Adapter registered within the Provider.
    :type adapter: BaseAdapter

    :param config: Configuration data of the Provider.
    :type config: Config

    :param authenticate: Object used to authenticate the current Client.
    :type authenticate: ClientAuthentication
    """

    __endpoint__: str = None

    def __init__(
        self, adapter: BaseAdapter, config: Config, authenticate: ClientAuthentication
    ) -> None:
        self.adapter = adapter
        self.config = config
        self.authenticate = authenticate

    @abc.abstractmethod
    async def __call__(self, request: Request) -> Response:
        """
        All endpoints are required to implement this method,
        since it MUST return a response back to the client.

        The type, status, headers and body of the response it returns,
        as well as its meaning and formatting have to be documented by
        the respective endpoint.

        This method **MUST NOT** raise **ANY** exception.

        If an error occurred during the processing of the request,
        it **MUST** be treated and its appropriate response status,
        headers and body **MUST** be correctly set
        to denote the type of exception that occured.

        Other than the previous requirements, the endpoint is free
        to use whatever tools, methods and workflows it wished.

        It is recommended to split the logic of the endpoint into small
        single-responsibility methods for better maintenance.

        :param request: Current request being processed.
        :type request: Request

        :return: Response containing all the necessary info to the client.
        :rtype: Response
        """



from psion.jose import JsonWebTokenClaims
from psion.jose.exceptions import InvalidJWTClaim


class IDToken(JsonWebTokenClaims):
    """
    Defines the claims of the ID Token.

    The parameters supported out-of-the-box by Guarani are described below.
    To get more information, please refer to the specification at
    `<https://openid.net/specs/openid-connect-core-1_0.html#IDToken>`_::

        * "iss" - Issuer of the response.
        * "sub" - ID of the Authenticated User.
        * "aud" - Audience of the ID Token. Its value is the ID of the Client.
        * "exp" - Expiration time of the ID Token.
        * "iat" - Time when the ID token was issued.
        * "auth_time" - Time when the User was authenticated.
        * "nonce" - String value used to associate a Client session with an ID Token,
            and to mitigate replay attacks.
        * "acr" - Authentication Context Class Reference.*
        * "amr" - Authentication Methods References.*
        * "azp" - Authorized party ID. Refers to the ID of the Client.*

        The following is a non-normative example of the Claims of an ID Token:

        {
            "iss": "https://server.example.com",
            "sub": "24400320",
            "aud": "s6BhdRkqt3",
            "nonce": "n-0S6_WzA2Mj",
            "exp": 1311281970,
            "iat": 1311280970,
            "auth_time": 1311280969,
            "acr": "urn:mace:incommon:iap:silver"
        }

        * Support in the future.
    """

    def _validate_default_claims(self, claims: dict) -> None:
        super()._validate_default_claims(claims)

        # Validate auth_time
        self._validate_auth_time(claims)

        # Validate nonce
        self._validate_nonce(claims)

    def _validate_auth_time(self, claims: dict) -> None:
        auth_time = claims.get("max_age")

        if claims.get("max_age") and not auth_time:
            raise InvalidJWTClaim('Invalid claim "auth_time".')

        if auth_time is not None and type(auth_time) is not int:
            raise InvalidJWTClaim('Invalid claim "auth_time".')

    def _validate_nonce(self, claims: dict) -> None:
        if nonce := self.options.get("nonce"):
            if nonce != claims.get("nonce"):
                raise InvalidJWTClaim('Invalid claim "nonce".')



import abc
import binascii

from psion.jose.exceptions import InvalidKey, InvalidSignature
from psion.jose.jwk import JsonWebKey
from psion.webtools import base64url_decode, base64url_encode


class JWSAlgorithm(abc.ABC):
    """
    Implementation of the Section 3 of RFC 7518.

    This class provides the expected method signatures
    that will be used throughout the package.

    All JWS Algorithms **MUST** inherit from this class and
    implement its methods.

    :cvar ``__algorithm__``: Name of the algorithm.
    :cvar ``__hash_name__``: Name of the hash function used by the algorithm.
    :cvar ``__key_type__``: Type of the key that the algorithm accepts.
    """

    __algorithm__: str = None
    __hash_name__: str = None
    __key_type__: str = None

    @classmethod
    def validate_key(cls, key: JsonWebKey):
        """
        Validates the provided key against the algorithm's
        specifications and restrictions.

        :param key: JWK to be validated.
        :type key: JsonWebKey

        :raises InvalidKey: The provided key is invalid.
        """

        if not isinstance(key, JsonWebKey):
            raise InvalidKey

        # pylint: disable=used-before-assignment
        if (alg := key.data.get("alg")) and alg != cls.__algorithm__:
            raise InvalidKey(
                f'This key is intended to be used by the algorithm "{alg}".'
            )

        if key.data.get("kty") != cls.__key_type__:
            raise InvalidKey(f'This algorithm only accepts "{cls.__key_type__}" keys.')

    @classmethod
    @abc.abstractmethod
    def sign(cls, data: bytes, key: JsonWebKey) -> bytes:
        """
        Signs the provided data using the provided key.

        :param data: Data to be signed.
        :type data: bytes

        :param key: JWK used to sign the data.
        :type key: JsonWebKey

        :return: URL Safe Base64 encoded signature of the data.
        :rtype: bytes
        """

    @classmethod
    @abc.abstractmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey) -> None:
        """
        Verifies if the data and signature provided match
        based on the provided Json Web Key.

        :param signature: Signature used in the verification.
            **MUST** be a URL Safe Base64 encoded bytes string.
        :type signature: bytes

        :param data: Data to be verified.
        :type data: bytes

        :param key: JWK used to verify the data.
        :type key: JsonWebKey

        :raises InvalidSignature: The signature and data do not match.
        """


class none(JWSAlgorithm):
    __algorithm__: str = "none"

    @classmethod
    def sign(cls, data: bytes, key: JsonWebKey = None) -> bytes:
        return b""

    @classmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey = None) -> None:
        pass


class _HMAC(JWSAlgorithm):
    __key_type__: str = "oct"

    @classmethod
    def sign(cls, data: bytes, key: JsonWebKey) -> bytes:
        cls.validate_key(key)
        signature = key.sign(data, cls.__hash_name__)
        return base64url_encode(signature)

    @classmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey):
        cls.validate_key(key)

        try:
            # Incorrect padding of the encoded signature.
            raw_signature = base64url_decode(signature)
        except binascii.Error:
            raise InvalidSignature

        key.verify(raw_signature, data, cls.__hash_name__)


class HS256(_HMAC):
    __algorithm__: str = "HS256"
    __hash_name__: str = "SHA256"


class HS384(_HMAC):
    __algorithm__: str = "HS384"
    __hash_name__: str = "SHA384"


class HS512(_HMAC):
    __algorithm__: str = "HS512"
    __hash_name__: str = "SHA512"


class _RSA_PKCS1v15(JWSAlgorithm):
    __key_type__: str = "RSA"
    __padding__: str = "PKCS1v15"

    @classmethod
    def sign(cls, data: bytes, key: JsonWebKey) -> bytes:
        cls.validate_key(key)
        signature = key.sign(data, cls.__hash_name__, rsa_padding=cls.__padding__)
        return base64url_encode(signature)

    @classmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey):
        cls.validate_key(key)

        try:
            # Incorrect padding of the encoded signature.
            raw_signature = base64url_decode(signature)
        except binascii.Error:
            raise InvalidSignature

        key.verify(raw_signature, data, cls.__hash_name__, rsa_padding=cls.__padding__)


class RS256(_RSA_PKCS1v15):
    __algorithm__: str = "RS256"
    __hash_name__: str = "SHA256"


class RS384(_RSA_PKCS1v15):
    __algorithm__: str = "RS384"
    __hash_name__: str = "SHA384"


class RS512(_RSA_PKCS1v15):
    __algorithm__: str = "RS512"
    __hash_name__: str = "SHA512"


class _EC(JWSAlgorithm):
    __curve__: str = None
    __key_type__: str = "EC"

    @classmethod
    def validate_key(cls, key: JsonWebKey):
        super(_EC, cls).validate_key(key)

        if key.data.get("crv") != cls.__curve__:
            raise InvalidKey(
                f'This algorithm only accepts the curve "{cls.__curve__}".'
            )

    @classmethod
    def sign(cls, data: bytes, key: JsonWebKey) -> bytes:
        cls.validate_key(key)
        signature = key.sign(data, cls.__hash_name__)
        return base64url_encode(signature)

    @classmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey):
        cls.validate_key(key)

        try:
            # Incorrect padding of the encoded signature.
            raw_signature = base64url_decode(signature)
        except binascii.Error:
            raise InvalidSignature

        key.verify(raw_signature, data, cls.__hash_name__)


class ES256(_EC):
    __algorithm__: str = "ES256"
    __curve__: str = "P-256"
    __hash_name__: str = "SHA256"


class ES384(_EC):
    __algorithm__: str = "ES384"
    __curve__: str = "P-384"
    __hash_name__: str = "SHA384"


class ES512(_EC):
    __algorithm__: str = "ES512"
    __curve__: str = "P-521"
    __hash_name__: str = "SHA512"


class _RSA_PSS(JWSAlgorithm):
    __key_type__: str = "RSA"
    __padding__: str = "PSS"

    @classmethod
    def sign(cls, data: bytes, key: JsonWebKey) -> bytes:
        cls.validate_key(key)
        signature = key.sign(data, cls.__hash_name__, rsa_padding=cls.__padding__)
        return base64url_encode(signature)

    @classmethod
    def verify(cls, signature: bytes, data: bytes, key: JsonWebKey):
        cls.validate_key(key)

        try:
            # Incorrect padding of the encoded signature.
            raw_signature = base64url_decode(signature)
        except binascii.Error:
            raise InvalidSignature

        key.verify(raw_signature, data, cls.__hash_name__, rsa_padding=cls.__padding__)


class PS256(_RSA_PSS):
    __algorithm__: str = "PS256"
    __hash_name__: str = "SHA256"


class PS384(_RSA_PSS):
    __algorithm__: str = "PS384"
    __hash_name__: str = "SHA384"


class PS512(_RSA_PSS):
    __algorithm__: str = "PS512"
    __hash_name__: str = "SHA512"



from __future__ import annotations
from typing import Any

from psion.oauth2.exceptions import (
    InvalidRequest,
    OAuth2Error,
    UnauthorizedClient,
    UnsupportedGrantType,
)
from psion.oauth2.grants import BaseGrant
from psion.oauth2.models import JSONResponse, Request

from .base import BaseEndpoint


class TokenEndpoint(BaseEndpoint):
    """
    Endpoint used by the client to exchange an authorization grant,
    or its own credentials for an access token that will be used by
    the client to act on behalf of the Resource Owner.

    This endpoint requires some kind of `Client Authentication`.
    The methods used to authenticate or validate the Client **MUST**
    be defined by the `Grants` that make use of this endpoint
    via the Grant attribute :attr:`__authenticate_methods__`.
    """

    __endpoint__: str = "token"

    _headers = {"Cache-Control": "no-store", "Pragma": "no-cache"}

    def __init__(self, grants: list[BaseGrant], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.grants = grants

    async def __call__(self, request: Request) -> JSONResponse:
        """
        Creates a Token Response via a JSON Response.

        This endpoint is responsible for issuing Tokens to Clients
        that succeed to authenticate within the Authorization Server
        and has the necessary consent of the Resource Owner.

        This endpoint is to be used by Grants that have a Token Workflow,
        and it will return a JSON object as a result.

        If the Client fails to authenticate within the Authorization Server,
        does not have the consent of the Resource Owner or provides invalid
        or insufficient request parameters, it will receive a `400 Bad Request`
        Error Response with a JSON object describing the error.

        If the flow succeeds, the Client will then receive its Token
        in a JSON object containing the Access Token, the Token Type,
        the Lifespan of the Access Token and an optional Refresh Token,
        as well as an optional Scope parameter if the granted scopes
        differ from the requested ones.

        :param request: Current request being processed.
        :type request: Request

        :return: Token Response and its metadata.
        :rtype: JSONResponse
        """

        try:
            data = request.data

            if not data:
                raise InvalidRequest(description="Missing request parameters.")

            grant = self._validate_grant_type(data.pop("grant_type", None))
            client = await self.authenticate(request, grant.__authentication_methods__)

            if grant.__grant_type__ not in client.get_grant_types():
                raise UnauthorizedClient

            response = await grant.token(data, client)
            return JSONResponse(200, self._headers, response)
        except OAuth2Error as exc:
            headers = exc.headers
            headers.update(self._headers)
            return JSONResponse(400, headers, exc.dump())

    def _validate_grant_type(self, grant_type: str) -> BaseGrant:
        """
        Validates the requested `grant_type` against the set
        of registered Grants of the Provider.

        :param grant_type: Response type to be validated.
        :type grant_type: str

        :raises InvalidRequest: The `grant_type` is missing or invalid.
        :raises UnsupportedGrantType: The Provider does not support
            the requested `grant_type` as a Token Grant.

        :return: Grant that represents the requested `grant_type`.
        :rtype: BaseGrant
        """

        if not grant_type or not isinstance(grant_type, str):
            raise InvalidRequest(description='Invalid parameter "grant_type".')

        for grant in self.grants:
            if grant.__grant_type__ == grant_type:
                return grant
        else:
            raise UnsupportedGrantType(
                description=f'Unsupported grant_type "{grant_type}".'
            )



from psion.oauth2.exceptions import InvalidClient
from psion.oauth2.mixins import ClientMixin
from psion.oauth2.models import Request
from psion.webtools import get_basic_authorization

from .base import BaseAuthentication


class ClientSecretBasic(BaseAuthentication):
    """
    Implements the Client Authentication via the Basic Authentication workflow.

    If this workflow is enabled, it will look at the Authorization header
    for a scheme similar to the following::

        Basic Y2xpZW50MTpjbGllbnQxc2VjcmV0

    This scheme denotes the type of the flow, which in this case is `Basic`,
    and the Client Credentials, that is a Base64 encoded string that contains
    the Credentials in the format `client_id:client_secret`.
    """

    __method__: str = "client_secret_basic"

    async def authenticate(self, request: Request) -> ClientMixin:
        client_id, client_secret = get_basic_authorization(request.headers)

        if not client_id or not client_secret:
            return None

        client = await self.find_client(client_id)
        headers = {"WWW-Authenticate": "Basic"}

        if not client:
            raise InvalidClient("Client not found.", headers=headers)

        if client.get_client_secret() != client_secret:
            raise InvalidClient("Mismatching Client Secret.", headers=headers)

        if client.get_token_endpoint_auth_method() != self.__method__:
            raise InvalidClient(
                f'This Client is not allowed to use the method "{self.__method__}".',
                headers=headers,
            )

        return client



from __future__ import annotations

from typing import Optional, Union

from psion.oauth2.mixins import (
    AuthorizationCodeMixin,
    ClientMixin,
    RefreshTokenMixin,
    UserMixin,
)


class BaseAdapter:
    """
    Base Adapter class that contains common methods used throughout Guarani.

    These methods are used by multiple authentication methods, endpoints and/or grants,
    and therefore, to respect the DRY principle, they are defined in this class.

    The application **MUST** provide a concrete implementation of the methods
    defined in this class.
    """

    async def find_user(self, user_id: str) -> UserMixin:
        """
        Searches for a user in the application's storage
        and returns it if it succeeds, otherwise returns None.

        :param user_id: ID of the user being searched.
        :type user_id: str

        :return: User based on the provided ID.
        :rtype: UserMixin
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def find_client(self, client_id: str) -> ClientMixin:
        """
        Searches for a client in the application's storage
        and returns it if it succeeds, otherwise returns None.

        :param client_id: ID of the client being searched.
        :type client_id: str

        :return: Client based on the provided ID.
        :rtype: ClientMixin
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def save_authorization_code(
        self, code: str, data: dict, client: ClientMixin, user: UserMixin
    ) -> None:
        """
        Binds the `Authorization Code` to the `Client` and saves it for the Token part.

        It is **RECOMMENDED** that the application sets a lifetime for the code.

        :param code: Code to be associated with the Client.
        :type code: str

        :param data: Dictionary containing the data of the Authorization Request.
        :type data: dict

        :param client: Client being authorized.
        :type client: ClientMixin

        :param user: User granting authorization.
        :type user: UserMixin
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def get_authorization_code(self, code: str) -> AuthorizationCodeMixin:
        """
        Retrieves the data of an `Authorization Code`from
        the application's storage based on the provided `code`.

        :param code: Authorization Code to be fetched.
        :type code: str

        :return: Requested Authorization Code.
        :rtype: AuthorizationCodeMixin
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def delete_authorization_code(self, code: str) -> None:
        """
        Deletes the provided `Authorization Code` from the application's storage.

        :param code: Authorization Code to be deleted.
        :type code: str
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def create_access_token(
        self,
        client: ClientMixin,
        resource_owner: Union[ClientMixin, UserMixin],
        scopes: list[str],
    ) -> str:
        """
        Generates an `Access Token` that creates a tight coupling between the `Client`,
        the `Resource Owner` and the `Scopes` that the `Resource Owner`
        authorized the `Client` to access on its behalf.

        The structure of the `Access Token` is left undefined by this framework,
        but it is **RECOMMENDED** that the application uses `Json Web Token (JWT)`
        for the `Access Token`, followin the specifications at
        `JWT Profile for OAuth 2 Access Tokens
        <https://tools.ietf.org/html/draft-ietf-oauth-access-token-jwt>`_.

        The format of the `Access Token Response` is as follows, with values displayed
        as example only, not defining a format for any of the Tokens::

            {
                "access_token": "vlOa11kBoziWFBsQiUu59SjgHJbi7spU80Ew5xCTZ9UhZmWN",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token": "7eqGioGLs-O7ky3CgeAU87bfijRam6r5"
            }

        Since the `Refresh Token` is optional when the `Resource Owner`
        is separate from the `Client`, and not recommended
        when the `Client` is the `Resource Owner`, its presence
        in the final response is optional.

        :param client: Client that will use the issued Access Token.
        :type client: ClientMixin

        :param resource_owner: Resource Owner that the Client is acting on behalf of.
        :type resource_owner: Union[ClientMixin, UserMixin]

        :param scopes: Scopes of the Token granted by the Resource Owner to the Client.
        :type scopes: list[str]

        :return: Issued Access Token.
        :rtype: str
        """

        raise NotImplementedError("This method MUST be implemented.")

    async def get_userinfo(self, user: UserMixin, scopes: list[str]) -> dict:
        """
        Returns the information about the User based on the requested scopes.

        :param user: User subject of the Authentication Request.
        :type user: UserMixin

        :param scopes: Scopes requested by the Client.
        :type scopes: list[str]

        :return: Information about the User based on the requested scopes.
        :rtype: dict
        """

        raise NotImplementedError("Implement this method to support ID Tokens.")

    async def get_key_info(self) -> dict:
        """
        Returns the information necessary to create an ID Token.

        The dictionary returned **MUST** contain the following information::

            * "key": Key used to sign the ID Token.
            * "alg": Algorithm used to create the signature the ID Token.

        :return: Info to create an ID Token.
        :rtype: dict
        """

        raise NotImplementedError("Implement this method to support ID Tokens.")

    async def create_refresh_token(
        self, client: ClientMixin, user: UserMixin, scopes: list[str]
    ) -> str:
        """
        Generates a `Refresh Token` binding the `Client` and the `User`,
        together with the `Scopes` granted by the `User`, and sets its
        expiration to the value of the argument `expiration_in_days`.

        :param client: Client to whom the Refresh Token was issued to.
        :type client: ClientMixin

        :param user: User represented by the Client.
        :type user: UserMixin

        :param scopes: Scopes of the next Access Token issued by this Refresh Token.
        :type scopes: list[str]

        :return: Issued Refresh Token.
        :rtype: str
        """

        raise NotImplementedError("Implement this method to support Refresh Tokens.")

    async def get_refresh_token(self, refresh_token: str) -> RefreshTokenMixin:
        """
        Retrieves the data of a `Refresh Token` from the application's storage
        based on the provided `refresh_token`.

        :param refresh_token: Refresh Token to be fetched.
        :type refresh_token: str

        :return: Requested Refresh Token.
        :rtype: RefreshTokenMixin
        """

        raise NotImplementedError("Implement this method to support Refresh Tokens.")

    async def revoke_token(
        self, client: ClientMixin, token: str, token_type_hint: Optional[str] = None
    ) -> None:
        """
        Searches the application's storage for the provided token and revokes it.

        :param client: Client requesting a token revocation.
        :type client: ClientMixin

        :param token: Token to be revoked.
        :type token: str

        :param token_type_hint: Tells the server the type of the token, defaults to None.
        :type token_type_hint: str, optional
        """

        raise NotImplementedError("Implement this method to support Token Revocation.")



from __future__ import annotations

import abc
import base64
import enum
import os
import secrets
import warnings
from typing import Any, Literal, Optional, Sequence, Union

from cryptography.exceptions import InvalidSignature as BaseInvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac, serialization
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa

from psion.jose.exceptions import InvalidKey, InvalidSignature, UnsupportedParsingMethod
from psion.webtools import (
    FullDict,
    b64_to_int,
    base64url_decode,
    base64url_encode,
    int_to_b64,
    to_bytes,
    to_string,
)


class ALGORITHMS(enum.Enum):
    """
    Enumeration of the supported Hash Algorithms.
    """

    SHA256 = hashes.SHA256()
    SHA384 = hashes.SHA384()
    SHA512 = hashes.SHA512()


AlgType = Union[str, ALGORITHMS]
RSAPadding = Literal["PKCS1v15", "PSS"]

EC_PRIVATE = ec.EllipticCurvePrivateKeyWithSerialization
EC_PUBLIC = ec.EllipticCurvePublicKey
RSA_PRIVATE = rsa.RSAPrivateKeyWithSerialization
RSA_PUBLIC = rsa.RSAPublicKey


def _get_alg(alg: AlgType) -> hashes.HashAlgorithm:
    """
    Auxiliar function used to translate a string into an ALGORITHMS item.

    :param alg: Name of the algorithm to be retrieved.
    :type alg: AlgType

    :return: Requested algorithm.
    """

    if isinstance(alg, str):
        return ALGORITHMS[alg].value

    return alg.value


def _parse_raw(path_or_secret: os.PathLike | bytes) -> bytes:
    try:
        with open(path_or_secret, "rb") as f:
            return f.read()
    except (FileNotFoundError, IsADirectoryError):
        return to_bytes(path_or_secret)


class JWKAlgorithm(abc.ABC):
    """
    Implementation of the Section 6 of RFC 7518.

    This base class provides the signatures of the methods used by the Json Web Key
    implementation regarding to the Key Algorithms usage.

    Implementations of custom Key Algorithms **MUST** inherit from this base class
    in order to be compliant with the rest of the library.

    :cvar ``__allowed_attributes__``: Defines the group of attributes
        supported by the Key Algorithm.
    :cvar kty: The name of the Key Algorithm.
    """

    __allowed_attributes__: Sequence[str] = None

    kty: str = None

    @classmethod
    @abc.abstractmethod
    def generate(cls, **kwargs: Any) -> JWKAlgorithm:
        """
        Generates a key on the fly based on the provided arguments.

        :return: Generated key as JWKAlgorithm.
        :rtype: JWKAlgorithm
        """

    @classmethod
    @abc.abstractmethod
    def parse(
        cls,
        path_or_secret: os.PathLike | bytes,
        password: bytes = None,
        format: str = "pem",
    ) -> JWKAlgorithm:
        """
        Parses a raw key into a JWKAlgorithm.

        A raw symmetric key is simply its bytes string.
        A raw asymmetric key would be a PEM encoded key data.

        :param path_or_secret: The path of the raw key or its bytes representation.
        :type path_or_secret: os.PathLike | bytes

        :param password: Password used to decrypt the raw key, defaults to None.
        :type password: bytes, optional

        :param format: The format of the raw key, defaults to `pem`.
            If `pem`, assumes it is Base64 Encoded.
            If `der`, assumes it is a regular sequence of bytes.
        :type format: str, optional

        :raises UnsupportedParsingMethod: Method not supported.
        :raises InvalidKey: The raw key type is different from the class.

        :return: Parsed key as a JWKAlgorithm.
        :rtype: JWKAlgorithm
        """

    @abc.abstractmethod
    def dump(self, public: bool = True) -> dict[str, Any]:
        """
        Returns a JSON-ready dictionary representation of the key's parameters.

        :param public: Dumps the public info of the key, defaults to True.
        :type public: bool, optional

        :return: Key in dict format.
        :rtype: dict
        """

    @abc.abstractmethod
    def export(self, public: bool = False) -> bytes:
        """
        Exports the key in PEM format if asymmetric, or Base64 if symmetric.

        :param public: Exports the public key, defaults to False.
        :type public: bool, optional

        :return: Base64/PEM encoded key data.
        :rtype: bytes
        """

    @abc.abstractmethod
    def sign(self, data: bytes, algorithm: AlgType, **kwargs: Any) -> bytes:
        """
        Creates a digital signature of the provided data.

        :param data: Data to be signed.
        :type data: bytes

        :param algorithm: Hash algorithm used to sign the data.
        :type algorithm: AlgType

        :return: Signature of the provided data.
        :rtype: bytes
        """

    @abc.abstractmethod
    def verify(
        self, signature: bytes, data: bytes, algorithm: AlgType, **kwargs: Any
    ) -> None:
        """
        Verifies the provided digital signature against the provided data.

        :param signature: Digital signature to be verified.
        :type signature: bytes

        :param data: Data used to verify the signature.
        :type data: bytes

        :param algorithm: Hash algorithm used to verify the signature.
        :type algorithm: AlgType

        :raises InvalidSignature: The signature does not match the data.
        """


class ECKey(JWKAlgorithm):
    """
    Implementation of the Elliptic Curve Asymmetric Key Algorithm.

    The standard curves are: "P-256", "P-384", "P-521".

    It is possible to add different curves, but they should be implemented
    by the application for a good support.

    :param crv: Name of the curve of the key.
    :type crv: str

    :param x: X coordinate of the Elliptic Curve.
    :type x: str

    :param y: Y coordinate of the Elliptic Curve.
    :type y: str

    :param d: Private value. **MANDATORY** if it is a private key.
    :type d: str, optional
    """

    __allowed_attributes__ = frozenset(("crv", "x", "y", "d"))

    kty: str = "EC"

    CURVES: dict[str, ec.EllipticCurve] = {
        "P-256": ec.SECP256R1(),
        "P-384": ec.SECP384R1(),
        "P-521": ec.SECP521R1(),
    }

    CURVES_NAMES: dict[str, str] = {
        ec.SECP256R1.name: "P-256",
        ec.SECP384R1.name: "P-384",
        ec.SECP521R1.name: "P-521",
    }

    def __init__(
        self, crv: str, x: str, y: str, d: Optional[str] = None, **ignore
    ) -> None:
        if crv not in self.CURVES.keys():
            raise InvalidKey(f'Unknown curve: "{crv}".')

        self._private = None
        self._public = None

        curve = self.CURVES[crv]
        x_coord = b64_to_int(x)
        y_coord = b64_to_int(y)

        public = ec.EllipticCurvePublicNumbers(x_coord, y_coord, curve)
        self._public: EC_PUBLIC = public.public_key(default_backend())

        if d:
            private_value = b64_to_int(d)
            private = ec.EllipticCurvePrivateNumbers(private_value, public)
            self._private: EC_PRIVATE = private.private_key(default_backend())

    @classmethod
    def generate(cls, curve: str) -> ECKey:
        """
        Generates a key on the fly based on the provided curve name.

        :param curve: Curve used to generate the key.
        :type curve: str

        :raises InvalidKey: Invalid parameters for the key.

        :return: Generated key as ECKey.
        :rtype: ECKey
        """

        if not (crv := cls.CURVES.get(curve)):
            raise InvalidKey(f'Unknown curve "{curve}".')

        key: EC_PRIVATE = ec.generate_private_key(crv, default_backend())

        private = key.private_numbers()
        public = key.public_key().public_numbers()

        return cls(
            crv=curve,
            x=to_string(int_to_b64(public.x)),
            y=to_string(int_to_b64(public.y)),
            d=to_string(int_to_b64(private.private_value)),
        )

    @classmethod
    def parse(
        cls,
        path_or_secret: os.PathLike | bytes,
        password: bytes = None,
        format: str = "pem",
    ) -> ECKey:
        """
        Parses a raw key into an ECKey.

        :param path_or_secret: The path of the raw key or its bytes representation.
        :type path_or_secret: os.PathLike | bytes

        :param password: Password used to decrypt the raw key, defaults to None.
        :type password: bytes, optional

        :param format: The format of the raw key, defaults to `pem`.
            If `pem`, assumes it is PEM Encoded.
            If `der`, assumes it is a regular sequence of bytes.
        :type format: str, optional

        :raises UnsupportedParsingMethod: Method not supported.
        :raises InvalidKey: The raw key type is different from the class.

        :return: Parsed key as ECKey.
        :rtype: ECKey
        """

        raw = _parse_raw(path_or_secret)

        if format == "pem":
            if b"PRIVATE" in raw:
                key: EC_PRIVATE = serialization.load_pem_private_key(
                    raw, password, default_backend()
                )

                if not isinstance(key, ec.EllipticCurvePrivateKey):
                    raise InvalidKey(
                        "The raw key is not an Elliptic Curve Private Key."
                    )

                private = key.private_numbers()
                public = key.public_key().public_numbers()

                return cls(
                    crv=cls.CURVES_NAMES.get(public.curve.name),
                    x=to_string(int_to_b64(public.x)),
                    y=to_string(int_to_b64(public.y)),
                    d=to_string(int_to_b64(private.private_value)),
                )

            if b"PUBLIC" in raw:
                key: EC_PUBLIC = serialization.load_pem_public_key(
                    raw, default_backend()
                )

                if not isinstance(key, ec.EllipticCurvePublicKey):
                    raise InvalidKey("The raw key is not an Elliptic Curve Public Key.")

                public = key.public_numbers()

                return cls(
                    crv=cls.CURVES_NAMES.get(public.curve.name),
                    x=to_string(int_to_b64(public.x)),
                    y=to_string(int_to_b64(public.y)),
                )

            raise InvalidKey("Unknown raw key format for Elliptic Curve.")

        raise UnsupportedParsingMethod

    def dump(self, public: bool = True) -> dict:
        """
        Returns a JSON-ready dictionary representation of the key's parameters.

        :param public: Dumps the public info of the key, defaults to True.
        :type public: bool, optional

        :return: Key in dict format.
        :rtype: dict
        """

        numbers = self._public.public_numbers()
        data = {
            "kty": self.kty,
            "crv": self.CURVES_NAMES[numbers.curve.name],
            "x": to_string(int_to_b64(numbers.x)),
            "y": to_string(int_to_b64(numbers.y)),
        }

        if public:
            return data

        if not self._private:
            raise RuntimeError("This key is not a private key.")

        private_value = self._private.private_numbers().private_value
        return FullDict(data, d=to_string(int_to_b64(private_value)))

    def export(self, public: bool = False) -> bytes:
        """
        Exports the key in PEM format.

        :param public: Exports the public key, defaults to False.
        :type public: bool, optional

        :return: PEM encoded key data.
        :rtype: bytes
        """

        if not public:
            if self._private:
                return self._private.private_bytes(
                    serialization.Encoding.PEM,
                    serialization.PrivateFormat.TraditionalOpenSSL,
                    serialization.NoEncryption(),
                )

            raise InvalidKey("No private key found.")

        return self._public.public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def sign(self, data: bytes, algorithm: AlgType) -> bytes:
        """
        Returns a signature of the provided data.

        :param data: Data to be signed.
        :type data: bytes

        :param algorithm: Hash algorithm used to sign the provided data.
        :type algorithm: AlgType

        :return: Signature of the provided data.
        :rtype: bytes
        """

        if not self._private:
            raise InvalidKey("Cannot sign with a public key.")

        return self._private.sign(data, ec.ECDSA(_get_alg(algorithm)))

    def verify(self, signature: bytes, data: bytes, algorithm: AlgType) -> None:
        """
        Compares the provided signature against the provided data.

        :param signature: Signature to be compared.
        :type signature: bytes

        :param data: Data to be compared against.
        :type data: bytes

        :param algorithm: Hash algorithm used to validate the data and the signature.
        :type algorithm: AlgType

        :raises InvalidSignature: The signature and data do not match.
        """

        try:
            self._public.verify(signature, data, ec.ECDSA(_get_alg(algorithm)))
        except BaseInvalidSignature:
            raise InvalidSignature


class OCTKey(JWKAlgorithm):
    """
    Implementation of a symmetric key.

    In this implementation, the same secret is used to perform all of the operations.

    It is **NOT RECOMMENDED** to disclose this type of key in a Json Web Key Set (JWKS),
    since it **COULD** lead to security issues.

    :param k: The secret used in this key. It is assumed that the secret will be
        a Base64UrlEncoded string and, thus, it will perform the necessary conversions.
    :type k: str
    """

    __allowed_attributes__ = frozenset(("k"))

    kty: str = "oct"

    def __init__(self, k: str, **ignore) -> None:
        self._secret = base64url_decode(to_bytes(k))

    @classmethod
    def generate(cls, size: int = 32) -> OCTKey:
        """
        Generates a secure random bytes sequence based on the provided size.

        :param size: Size of the secret in bytes, defaults to 32.
        :type size: int, optional

        :raises InvalidKey: Invalid parameters for the key.

        :return: Instance of an OCTKey.
        :rtype: OCTKey
        """

        if size < 32:
            raise InvalidKey("Size is too short. MUST be AT LEAST 32 bytes.")

        secret = base64url_encode(secrets.token_bytes(size))

        return cls(k=to_string(secret))

    @classmethod
    def parse(
        cls,
        path_or_secret: os.PathLike | bytes,
        password: bytes = None,
        format: str = "pem",
    ) -> OCTKey:
        """
        Parses a raw secret into an OCTKey.

        :param path_or_secret: The path of the raw key or its bytes representation.
        :type path_or_secret: os.PathLike | bytes

        :param password: Password used to decrypt the raw key, defaults to None.
        :type password: bytes, optional

        :param format: The format of the raw key, defaults to `pem`.
            If `pem`, assumes it is Base64 Encoded.
            If `der`, assumes it is a regular sequence of bytes.
        :type format: str, optional

        :raises UnsupportedParsingMethod: Method not supported.
        :raises InvalidKey: The raw key type is different from the class.

        :return: Parsed key as OCTKey.
        :rtype: OCTKey
        """

        raw = _parse_raw(path_or_secret)

        if format == "pem":
            invalid_strings = [
                b"-----BEGIN CERTIFICATE-----",
                b"-----BEGIN PRIVATE KEY-----",
                b"-----BEGIN RSA PRIVATE KEY-----",
                b"-----BEGIN EC PRIVATE KEY-----",
                b"-----BEGIN PUBLIC KEY-----",
                b"-----BEGIN RSA PUBLIC KEY-----",
                b"-----BEGIN EC PUBLIC KEY-----",
                b"ssh-rsa",
            ]

            if any(string in raw for string in invalid_strings):
                raise InvalidKey(
                    "The raw key is an asymmetric key or X.509 Certificate "
                    "and CANNOT be used as a symmetric key."
                )

            data = to_string(base64url_encode(base64.b64decode(raw)))
            return cls(data)

        if format == "der":
            data = to_string(base64url_encode(raw))
            return cls(data)

        raise UnsupportedParsingMethod

    def dump(self, public: bool = True) -> dict:
        """
        Returns a JSON-ready dictionary representation of the key's parameters.

        :param public: Dumps the public info of the key, defaults to True.
        :type public: bool, optional

        :return: Key in dict format.
        :rtype: dict
        """

        if public:
            warnings.warn("Secret keys fo not have public info.", RuntimeWarning)

        return {"kty": self.kty, "k": to_string(base64url_encode(self._secret))}

    def export(self, public: bool = False) -> bytes:
        """
        Exports the key in Base64 format.

        :param public: Exports the public info, defaults to False.
        :type public: bool, optional

        :return: Base64 encoded key data.
        :rtype: bytes
        """

        if public:
            warnings.warn("Secret keys do not have public info.", RuntimeWarning)

        return base64.b64encode(base64url_decode(to_bytes(self._secret)))

    def sign(self, data: bytes, algorithm: AlgType) -> bytes:
        """
        Returns a signature of the provided data.

        :param data: Data to be signed.
        :type data: bytes

        :param algorithm: Hash algorithm used to sign the provided data.
        :type algorithm: AlgType

        :return: Signature of the provided data.
        :rtype: bytes
        """

        signature = hmac.HMAC(self._secret, _get_alg(algorithm), default_backend())
        signature.update(data)
        return signature.finalize()

    def verify(self, signature: bytes, data: bytes, algorithm: AlgType) -> None:
        """
        Compares the provided signature against the provided data.

        :param signature: Signature to be compared.
        :type signature: bytes

        :param data: Data to be compared against.
        :type data: bytes

        :param algorithm: Hash algorithm used to validate the data and the signature.
        :type algorithm: AlgType

        :raises InvalidSignature: The signature and data do not match.
        """

        try:
            message = hmac.HMAC(self._secret, _get_alg(algorithm), default_backend())
            message.update(data)
            message.verify(signature)
        except BaseInvalidSignature:
            raise InvalidSignature


class RSAKey(JWKAlgorithm):
    """
    Implementation of the RSA Asymmetric Key Algorithm.

    This Key Algorithm requires the usage of a padding. Because of this requirement,
    the :meth:`sign` and :meth:`verify` methods require one more argument to be provided.

    :param n: Modulus of the key.
    :type n: str

    :param e: Public exponent.
    :type e: str

    :param d: Private exponent. **MANDATORY** if it is a private key.
    :type d: str, optional

    :param p: First prime coefficient.
    :type p: str, optional

    :param q: Second prime coefficient.
    :type q: str, optional

    :param dp: First prime CRT exponent.
    :type dp: str, optional

    :param dq: Second prime CRT exponent.
    :type dq: str, optional

    :param qi: First CRT coefficient.
    :type qi: str, optional
    """

    __allowed_attributes__ = frozenset(("n", "e", "d", "p", "q", "dp", "dq", "qi"))

    kty: str = "RSA"

    def __init__(
        self,
        n: str,
        e: str,
        d: Optional[str] = None,
        p: Optional[str] = None,
        q: Optional[str] = None,
        dp: Optional[str] = None,
        dq: Optional[str] = None,
        qi: Optional[str] = None,
        **ignore,
    ) -> None:
        self._private = None
        self._public = None

        modulus = b64_to_int(n)
        public_exponent = b64_to_int(e)

        public = rsa.RSAPublicNumbers(public_exponent, modulus)
        self._public: RSA_PUBLIC = public.public_key(default_backend())

        if d:
            private_exponent = b64_to_int(d)
            first_prime = b64_to_int(p)
            second_prime = b64_to_int(q)
            first_prime_crt = b64_to_int(dp)
            second_prime_crt = b64_to_int(dq)
            crt_coefficient = b64_to_int(qi)

            if not first_prime or not second_prime:
                first_prime, second_prime = rsa.rsa_recover_prime_factors(
                    modulus, public_exponent, private_exponent
                )

            if not first_prime_crt:
                first_prime_crt = rsa.rsa_crt_dmp1(private_exponent, first_prime)

            if not second_prime_crt:
                second_prime_crt = rsa.rsa_crt_dmq1(private_exponent, second_prime)

            if not crt_coefficient:
                crt_coefficient = rsa.rsa_crt_iqmp(first_prime, second_prime)

            private = rsa.RSAPrivateNumbers(
                first_prime,
                second_prime,
                private_exponent,
                first_prime_crt,
                second_prime_crt,
                crt_coefficient,
                public,
            )

            self._private: RSA_PRIVATE = private.private_key(default_backend())

    @classmethod
    def generate(cls, size: int = 2048) -> RSAKey:
        """
        Generates a key on the fly based on the provided module size.

        :param size: Size of the modulus in bits, defaults to 2048.
        :type size: int, optional

        :raises InvalidKey: Invalid parameters for the key.

        :return: Generated key as RSAKey.
        :rtype: RSAKey
        """

        if size < 512:
            raise InvalidKey("Size is too short. Must be AT LEAST 512 bits.")

        key = rsa.generate_private_key(65537, size, default_backend())

        private = key.private_numbers()
        public = key.public_key().public_numbers()

        return cls(
            n=to_string(int_to_b64(public.n)),
            e=to_string(int_to_b64(public.e)),
            d=to_string(int_to_b64(private.d)),
            p=to_string(int_to_b64(private.p)),
            q=to_string(int_to_b64(private.q)),
            dp=to_string(int_to_b64(private.dmp1)),
            dq=to_string(int_to_b64(private.dmq1)),
            qi=to_string(int_to_b64(private.iqmp)),
        )

    @classmethod
    def parse(
        cls,
        path_or_secret: os.PathLike | bytes,
        password: bytes = None,
        format: str = "pem",
    ) -> RSAKey:
        """
        Parses a raw key into an RSAKey.

        :param path_or_secret: The path of the raw key or its bytes representation.
        :type path_or_secret: os.PathLike | bytes

        :param password: Password used to decrypt the raw key, defaults to None.
        :type password: bytes, optional

        :param format: The format of the raw key, defaults to `pem`.
            If `pem`, assumes it is PEM Encoded.
            If `der`, assumes it is a regular sequence of bytes.
        :type format: str, optional

        :raises UnsupportedParsingMethod: Method not supported.
        :raises InvalidKey: The raw key type is different from the class.

        :return: Parsed key as RSAKey.
        :rtype: RSAKey
        """

        raw = _parse_raw(path_or_secret)

        if format == "pem":
            if b"PRIVATE" in raw:
                key: RSA_PRIVATE = serialization.load_pem_private_key(
                    raw, password, default_backend()
                )

                if not isinstance(key, rsa.RSAPrivateKey):
                    raise InvalidKey("The raw key is not an RSA Private Key.")

                private = key.private_numbers()
                public = key.public_key().public_numbers()

                return cls(
                    n=to_string(int_to_b64(public.n)),
                    e=to_string(int_to_b64(public.e)),
                    d=to_string(int_to_b64(private.d)),
                    p=to_string(int_to_b64(private.p)),
                    q=to_string(int_to_b64(private.q)),
                    dp=to_string(int_to_b64(private.dmp1)),
                    dq=to_string(int_to_b64(private.dmq1)),
                    qi=to_string(int_to_b64(private.iqmp)),
                )

            if b"PUBLIC" in raw:
                key: RSA_PUBLIC = serialization.load_pem_public_key(
                    raw, default_backend()
                )

                if not isinstance(key, rsa.RSAPublicKey):
                    raise InvalidKey("The raw key is not an RSA Public Key.")

                public = key.public_numbers()

                return cls(
                    n=to_string(int_to_b64(public.n)), e=to_string(int_to_b64(public.e))
                )

            raise InvalidKey("Unknown raw key format for RSA.")

        raise UnsupportedParsingMethod

    def dump(self, public: bool = True) -> dict:
        """
        Returns a JSON-ready dictionary representation of the key's parameters.

        :param public: Dumps the public info of the key, defaults to True.
        :type public: bool, optional

        :return: Key in dict format.
        :rtype: dict
        """

        numbers = self._public.public_numbers()
        data = {
            "kty": self.kty,
            "n": to_string(int_to_b64(numbers.n)),
            "e": to_string(int_to_b64(numbers.e)),
        }

        if public:
            return data

        if not self._private:
            raise RuntimeError("This key is not a private key.")

        private = self._private.private_numbers()

        return FullDict(
            data,
            d=to_string(int_to_b64(private.d)),
            p=to_string(int_to_b64(private.p)),
            q=to_string(int_to_b64(private.q)),
            dp=to_string(int_to_b64(private.dmp1)),
            dq=to_string(int_to_b64(private.dmq1)),
            qi=to_string(int_to_b64(private.iqmp)),
        )

    def export(self, public: bool = False) -> bytes:
        """
        Exports the key in PEM format.

        :param public: Exports the public key, defaults to False.
        :type public: bool, optional

        :return: PEM encoded key data.
        :rtype: bytes
        """

        if not public:
            if self._private:
                return self._private.private_bytes(
                    serialization.Encoding.PEM,
                    serialization.PrivateFormat.TraditionalOpenSSL,
                    serialization.NoEncryption(),
                )

            raise InvalidKey("No private key found.")

        return self._public.public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def sign(self, data: bytes, algorithm: AlgType, rsa_padding: RSAPadding) -> bytes:
        """
        Returns a signature of the provided data.

        :param data: Data to be signed.
        :type data: bytes

        :param algorithm: Hash algorithm used to sign the provided data.
        :type algorithm: AlgType

        :param rsa_padding: Padding used by the key.
        :type rsa_padding: RSAPadding

        :return: Signature of the provided data.
        :rtype: bytes
        """

        if not self._private:
            raise InvalidKey("Cannot sign with a public key.")

        alg = _get_alg(algorithm)

        if rsa_padding == "PKCS1v15":
            return self._private.sign(data, padding.PKCS1v15(), alg)

        if rsa_padding == "PSS":
            return self._private.sign(
                data, padding.PSS(padding.MGF1(alg), padding.PSS.MAX_LENGTH), alg
            )

        raise InvalidKey("Unsupported RSA padding.")

    def verify(
        self,
        signature: bytes,
        data: bytes,
        algorithm: AlgType,
        rsa_padding: RSAPadding,
    ) -> None:
        """
        Compares the provided signature against the provided data.

        :param signature: Signature to be compared.
        :type signature: bytes

        :param data: Data to be compared against.
        :type data: bytes

        :param algorithm: Hash algorithm used to validate the data and the signature.
        :type algorithm: AlgType

        :param rsa_padding: Padding used by the key.
        :type rsa_padding: RSAPadding

        :raises InvalidSignature: The signature and data do not match.
        """

        try:
            alg = _get_alg(algorithm)

            if rsa_padding == "PKCS1v15":
                return self._public.verify(signature, data, padding.PKCS1v15(), alg)

            if rsa_padding == "PSS":
                return self._public.verify(
                    signature,
                    data,
                    padding.PSS(padding.MGF1(alg), padding.PSS.MAX_LENGTH),
                    alg,
                )

            raise InvalidKey("Unsupported padding.")
        except BaseInvalidSignature:
            raise InvalidSignature



from __future__ import annotations

from typing import Optional

from psion.oauth2.exceptions import InvalidClient, OAuth2Error, UnsupportedTokenType
from psion.oauth2.models import JSONResponse, Request

from .base import BaseEndpoint


class RevocationEndpoint(BaseEndpoint):
    """
    Endpoint used by the `Client` to revoke a token in its possession.

    If the Client succeeds to authenticate but provides a token that was
    not issued to itself, the `Provider` **DOES NOT** revoke the token,
    since the Client is not authorized to operate the token.

    If the token is already invalid, does not exist within the Provider
    or is otherwise unknown or invalid, it is also considered "revoked".

    :cvar `__authentication_methods__`: Allowed Client Authentication methods.
    :cvar `__supported_tokens__`: Token types supported by the endpoint.
    """

    __endpoint__: str = "revocation"
    __authentication_methods__: list[str] = None
    __supported_tokens__: list[str] = ["access_token", "refresh_token"]

    async def __call__(self, request: Request) -> JSONResponse:
        """
        Revokes a previously issued Token.

        First it validates the `Revocation Request` of the `Client`
        by making sure the required parameter "token" is present,
        and that the `Client` can authenticate with the allowed
        authentication methods.

        From the specification at
        `<https://www.rfc-editor.org/rfc/rfc7009.html#section-2.1>`_::

            The client constructs the request by including the following
            parameters using the "application/x-www-form-urlencoded" format in
            the HTTP request entity-body:

            token REQUIRED. The token that the client wants to get revoked.

            token_type_hint OPTIONAL. A hint about the type of the token
                submitted for revocation. Clients MAY pass this parameter in
                order to help the authorization server to optimize the token
                lookup. If the server is unable to locate the token using
                the given hint, it MUST extend its search across all of its
                supported token types. An authorization server MAY ignore
                this parameter, particularly if it is able to detect the
                token type automatically. This specification defines two
                such values:

                * access_token: An access token as defined in [RFC6749],
                    Section 1.4

                * refresh_token: A refresh token as defined in [RFC6749],
                    Section 1.5

                Specific implementations, profiles, and extensions of this
                specification MAY define other values for this parameter
                using the registry defined in Section 4.1.2.

            For example, a client may request the revocation of a refresh token
            with the following request:

                POST /revoke HTTP/1.1
                Host: server.example.com
                Content-Type: application/x-www-form-urlencoded
                Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW

                token=45ghiukldjahdnhzdauz&token_type_hint=refresh_token

        It then returns an empty response with a HTTP Status 200 OK,
        signaling that the provided token has been revoked by the server.

        From the specification at
        `<https://www.rfc-editor.org/rfc/rfc7009.html#section-2.2>`_::

            The authorization server responds with HTTP status code 200 if the
            token has been revoked successfully or if the client submitted an
            invalid token.

            Note: invalid tokens do not cause an error response since the client
            cannot handle such an error in a reasonable way. Moreover, the
            purpose of the revocation request, invalidating the particular token,
            is already achieved.

            The content of the response body is ignored by the client as all
            necessary information is conveyed in the response code.

            An invalid token type hint value is ignored by the authorization
            server and does not influence the revocation response.

        This endpoint does not return any errors, except when the provided
        `token_type_hint` is not supported by the Provider.

        :raises UnsupportedTokenType: The provided token_type_hint is not supported.
        """

        try:
            client = await self.authenticate(request, self.__authentication_methods__)

            data = request.form()

            token: str = data.get("token")
            token_type_hint: Optional[str] = data.get("token_type_hint")

            if not token or not isinstance(token, str):
                return

            if token_type_hint:
                if token_type_hint not in self.__supported_tokens__:
                    raise UnsupportedTokenType

            await self.adapter.revoke_token(client, token, token_type_hint)

            return JSONResponse()
        except InvalidClient as exc:
            return JSONResponse(401, exc.headers, exc.dump())
        except OAuth2Error as exc:
            return JSONResponse(400, exc.headers, exc.dump())



from __future__ import annotations

from datetime import datetime

from psion.jose.jwk import JsonWebKey


class AuthorizationCodeMixin:
    """
    Defines the model of the `Authorization Code` used by this framework.

    The application's Authorization Code **MUST** inherit from this class
    and implement **ALL** the methods defined here.
    """

    def get_client_id(self) -> str:
        """
        Returns the `ID` of the `Client` that requested the `Authorization Grant`.

        :return: ID of the Client from this Authorization Code.
        :rtype: str
        """

        raise NotImplementedError

    def get_user_id(self) -> str:
        """
        Returns the `ID` of the `User` that issued the current `Authorization Grant`.

        :return: ID of the User from this Authorization Code.
        :rtype: str
        """

        raise NotImplementedError

    def get_redirect_uri(self) -> str:
        """
        Returns the `Redirect URI` of the current `Authorization Code`.

        :return: Redirect URI.
        :rtype: str
        """

        raise NotImplementedError

    def get_scopes(self) -> list[str]:
        """
        Returns the `Scopes` that were authorized by the `User` to the `Client`.

        :return: Authorized Scopes.
        :rtype: list[str]
        """

        raise NotImplementedError

    def get_code_challenge(self) -> str:
        """
        Returns the `Code Challenge` provided by the `Client`.

        :return: Code Challenge.
        :rtype: str
        """

        raise NotImplementedError

    def get_code_challenge_method(self) -> str:
        """
        Returns the `Code Challenge Method` used by the `Client`.

        :return: Code Challenge Method.
        :rtype: str
        """

        raise NotImplementedError

    def get_nonce(self) -> str:
        """
        Returns the value of the `nonce` provided by the `Client`
        in the `Authentication Request`.

        :return: Nonce value of the Client.
        :rtype: str
        """

        raise NotImplementedError

    def get_auth_time(self) -> int:
        """
        Returns the time of the authentication of the User.

        :return: Time of User authentication.
        :rtype: int
        """

        raise NotImplementedError

    def get_expiration(self) -> datetime:
        """
        Returns a datetime representing the expiration of the `Authorization Code`.

        :return: Time when the Authorization Code expires.
        :rtype: datetime
        """

        raise NotImplementedError


class ClientMixin:
    """
    Defines the model of the `Client` used by this framework.

    The application's Client **MUST** inherit from this class and implement
    **ALL** the methods defined here.
    """

    def get_client_id(self) -> str:
        """
        Returns the `ID` of the `Client`.

        :return: ID of the Client.
        :rtype: str
        """

        raise NotImplementedError

    def get_client_secret(self) -> str:
        """
        Returns the `Secret` of the `Client`.

        :return: Secret of the Client.
        :rtype: str
        """

        raise NotImplementedError

    def get_client_public_key(self, key_id: str) -> JsonWebKey:
        """
        Returns an instance of the Client's Public Key based on
        the provided Key ID or the default Public key.

        :param key_id: ID of the Public Key to be retrieved.
        :type key_id: str

        :return: Instance of the Client's (default) Public Key.
        :rtype: JsonWebKey
        """

        raise NotImplementedError

    def get_allowed_scopes(self, scopes: list[str]) -> list[str]:
        """
        Returns the `Scopes` that the `Client` is allowed to used
        based on the requested scopes.

        :param scopes: Requested scopes.
        :type scopes: list[str]

        :return: Scopes that the Client is allowed to request.
        :rtype: list[str]
        """

        raise NotImplementedError

    def get_redirect_uris(self) -> list[str]:
        """
        Returns a list of the `Redirect URIs` registered for the current `Client`.

        :return: Redirect URIs registered for the current Client.
        :rtype: list[str]
        """

        raise NotImplementedError

    def get_token_endpoint_auth_method(self) -> str:
        """
        Returns the `Token Endpoint Auth Method` of the `Client` stored in the database.

        :return: Token Endpoint Authentication Method of the Client.
        :rtype: str
        """

        raise NotImplementedError

    def get_grant_types(self) -> list[str]:
        """
        Returns a list of the `Grant Types` that the `Client` is allowed to request.

        :return: Grant Types that the Client is allowed to request.
        :rtype: list[str]
        """

        raise NotImplementedError

    def get_response_types(self) -> list[str]:
        """
        Returns a list of the `Response Types` that the `Client` is allowed to request.

        :return: Response Types that the Client is allowed to request.
        :rtype: list[str]
        """

        raise NotImplementedError


class RefreshTokenMixin:
    """
    Defines the model of the `Refresh Token` used by this framework.

    The application's Refresh Token **MUST** inherit from this class
    and implement **ALL** the methods defined here.
    """

    def get_refresh_token(self) -> str:
        """
        Returns the string that represents the `Refresh Token` object.

        :return: Refresh Token value.
        :rtype: str
        """

        raise NotImplementedError

    def get_client_id(self) -> str:
        """
        Returns the `ID` of the `Client` bound to the `Refresh Token`.

        :return: ID of the Client from this Refresh Token.
        :rtype: str
        """

        raise NotImplementedError

    def get_user_id(self) -> str:
        """
        Returns the `ID` of the `User` bound to the current `Refresh Token`.

        :return: ID of the User from this Refresh Token.
        :rtype: str
        """

        raise NotImplementedError

    def get_scopes(self) -> list[str]:
        """
        Returns the `Scopes` that were authorized by the `User` to the `Client`.

        :return: Authorized Scopes.
        :rtype: list[str]
        """

        raise NotImplementedError

    def get_expiration(self) -> datetime:
        """
        Returns a datetime representing the expiration of the `Refresh Token`.

        :return: Time when the Refresh Token expires.
        :rtype: datetime
        """

        raise NotImplementedError


class UserMixin:
    """
    Defines the model of the `User` used by this framework.

    The application's User **MUST** inherit from this class and implement
    **ALL** the methods defined here.
    """

    def get_user_id(self) -> str:
        """
        Returns the `ID` of the `User`.

        :return: ID of the User.
        :rtype: str
        """

        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['psion',
 'psion.jose',
 'psion.jose.jwa',
 'psion.oauth2',
 'psion.oauth2.authentication',
 'psion.oauth2.authentication.methods',
 'psion.oauth2.endpoints',
 'psion.oauth2.grants',
 'psion.oauth2.models',
 'psion.oauth2.providers',
 'psion.oidc',
 'psion.oidc.flows']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=3.3.1,<4.0.0']

setup_kwargs = {
    'name': 'psion',
    'version': '0.1.2',
    'description': 'Asynchronous WebAuth library for python.',
    'long_description': '# Project Psion\n\nThis library provides an implementation for asynchronous authentication and authorization of web applications.\nIt provides support for OAuth 2.1 and OpenID Connect, as well as JOSE.\n\nFor more details, please visit the documentation.\n\nAny doubts and suggestions can be sent to my [email](mailto:eduardorbr7@gmail.com).\nJust prepend the title with #Psion#, and I will try my best to answer.\n\n## JSON Object Signature and Encryption (JOSE)\n\nIt implements the following RFCs.\n\n1. [x] RFC 7515 - Json Web Signature (JWS)\n2. [ ] RFC 7516 - Json Web Encryption (JWE)\n3. [x] RFC 7517 - Json Web Key and Keyset (JWK)\n4. [x] RFC 7518 - Json Web Algorithms (JWA)\n5. [x] RFC 7519 - Json Web Token (JWT)\n\nFor more details on each feature, please visit the respective documentation.\n\n## OAuth 2.1\n\nIt implements the following RFCs and Specs.\n\n1. [x] Draft - The OAuth 2.1 Authorization Framework\n2. [x] RFC 7009 - OAuth 2.0 Token Revocation\n\n# License\n\nThis project is licensed under the MIT License.\n',
    'author': 'Eduardo Ribeiro Rezende',
    'author_email': 'eduardorbr7@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/revensky/psion',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from psion.oauth2.exceptions import InvalidRequest
from psion.oauth2.mixins import ClientMixin

from .base import BaseGrant


class ClientCredentialsGrant(BaseGrant):
    """
    Implementation of the Client Credentials Grant described in the
    `OAuth 2.1 Spec <https://tools.ietf.org/html/draft-parecki-oauth-v2-1#section-4.2>`_.

    This grant **DOES NOT** have an authorization grant step. Therefore, its usage
    on the authorization endpoint **WILL** result in an error.

    It **ONLY** allows `confidential` or `credentialed` clients to use it,
    since it relies on the authentication of the client based on its credentials
    to issue an access token. It also **DOES NOT** issue refresh tokens.
    """

    __authentication_methods__: list[str] = [
        "client_secret_basic",
        "client_secret_post",
    ]
    __grant_type__: str = "client_credentials"

    async def token(self, data: dict, client: ClientMixin) -> dict:
        """
        Validates the `Client's Credential` to authenticate the `Client`
        and validates if the `Client` is allowed to use this `Grant`.

        It then issues a new `Access Token` to the `Client`, without a `Refresh Token`.
        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.2.3>`_::

            If the access token request is valid and authorized, the
            authorization server issues an access token as described in
            Section 5.1. A refresh token SHOULD NOT be included. If the request
            failed client authentication or is invalid, the authorization server
            returns an error response as described in Section 5.2.

            An example successful response:

            HTTP/1.1 200 OK
            Content-Type: application/json
            Cache-Control: no-store
            Pragma: no-cache

            {
                "access_token": "2YotnFZFEjr1zCsicMWpAA",
                "token_type": "Bearer",
                "expires_in": 3600,
                "example_parameter": "example_value"
            }

        :param data: Data of the Token Request.
        :type data: dict

        :param client: Client requesting a new Token.
        :type client: ClientMixin

        :return: Access Token and its metadata.
        :rtype: dict
        """

        data = self._validate_token_request(data)

        self._validate_requested_scopes(data["scopes"], client)

        scopes = client.get_allowed_scopes(data["scopes"])

        token = await self.adapter.create_access_token(client, client, scopes)

        return self._create_token(
            token,
            self.config.token_lifespan,
            None,
            scopes if scopes != data["scopes"] else None,
        )

    def _validate_token_request(self, data: dict) -> dict:
        """
        Validates the incoming data from the `Client` to ensure
        that **ALL** the required parameters were provided.

        From the specification at
        `<https://tools.ietf.org/html/draft-parecki-oauth-v2-1-03#section-4.2.2>`_::

            The client makes a request to the token endpoint by adding the
            following parameters using the "application/x-www-form-urlencoded"
            format per Appendix B with a character encoding of UTF-8 in the HTTP
            request entity-body:

            "grant_type": REQUIRED. Value MUST be set to "client_credentials".

            "scope": OPTIONAL. The scope of the access request as described by
                Section 3.3.

            The client MUST authenticate with the authorization server as
            described in Section 3.2.1.
            For example, the client makes the following HTTP request using
            transport-layer security (with extra line breaks for display purposes
            only):

            POST /token HTTP/1.1
            Host: server.example.com
            Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW
            Content-Type: application/x-www-form-urlencoded

            grant_type=client_credentials

            The authorization server MUST authenticate the client.

        :param data: Data of the Token Request.
        :type data: dict

        :return: Validated and reformatted Token Request data.
        :rtype: dict
        """

        scope: str = data.pop("scope", None)

        if scope:
            if not scope or not isinstance(scope, str):
                raise InvalidRequest(description='Invalid parameter "scope".')

        return {"scopes": scope.split() if scope else None}



from datetime import datetime

from psion.jose import JsonWebToken
from psion.jose.jwk import JsonWebKey
from psion.oauth2.adapter import BaseAdapter
from psion.oauth2.config import Config
from psion.oauth2.exceptions import InvalidRequest
from psion.oauth2.mixins import ClientMixin, UserMixin
from psion.oidc.claims import IDToken
from psion.webtools import create_half_hash


class AuthorizationCodeFlow:
    """
    Hook representing the Authorization Flow of OpenID Connect.

    :param adapter: Registerd adapter of the Provider.
    :type adapter: BaseAdapter

    :param config: Registered configuration of the Provider.
    :type config: Config
    """

    def __init__(self, adapter: BaseAdapter, config: Config):
        self.adapter = adapter
        self.config = config

    async def authorization_request(self, data: dict) -> dict:
        if nonce := data.get("nonce"):
            if not nonce or not isinstance(nonce, str):
                raise InvalidRequest(description='Invalid parameter "nonce".')

            return {"nonce": nonce}

    # pylint: disable=unused-argument
    async def token_response(
        self, token: dict, data: dict, client: ClientMixin, user: UserMixin
    ) -> dict:
        if "openid" in token["scope"]:
            id_token = await self._generate_id_token(token, client, user)
            return {"id_token": id_token}

    async def _generate_id_token(
        self, token: dict, client: ClientMixin, user: UserMixin
    ) -> str:
        """
        Generates an ID Token containing the claims of the currently authenticated User
        based on the scopes requested by the Client.

        This method **MUST** only be used if the scope `openid` has been requested,
        otherwise, it is ignored and treated as a normal OAuth 2.1 Authorization Request.

        :param token: Token containing the Access Token and the scopes.
        :type token: dict

        :param client: Client requesting authorization.
        :type client: ClientMixin

        :param user: Currently authenticated User.
        :type user:

        :return: JWT encoded ID Token containing the claims
            requested by the Client about the User.
        :rtype: str
        """

        key_info = await self.adapter.get_key_info()
        userinfo = await self.adapter.get_userinfo(user, token["scope"].split())
        now = int(datetime.utcnow().timestamp())

        key: JsonWebKey = key_info["key"]

        claims = IDToken(
            {
                "iss": self.config.issuer,
                "aud": client.get_client_id(),
                "exp": now + self.config.id_token_lifespan,
                "iat": now,
                "at_hash": create_half_hash(token["access_token"], key_info["alg"]),
                **userinfo,
            }
        )
        id_token = JsonWebToken(
            claims,
            {"alg": key_info["alg"], "typ": "JWT", "kid": key.data["kid"]},
        )

        return id_token.encode(key)
